import json
import logging
import re
import shutil
import sqlite3
import subprocess
import threading
import uuid
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path, PurePosixPath
from typing import IO, Dict, Optional, Tuple, Union
from urllib.parse import quote

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import FileResponse, StreamingResponse
from pydantic import BaseModel, Field

from ....core.config import load_repo_config
from ....core.file_chat_keys import ticket_chat_scope
from ....core.flows import (
    FlowController,
    FlowDefinition,
    FlowEventType,
    FlowRunRecord,
    FlowRunStatus,
    FlowStore,
)
from ....core.flows.reconciler import reconcile_flow_run
from ....core.flows.ux_helpers import (
    bootstrap_check as ux_bootstrap_check,
)
from ....core.flows.ux_helpers import (
    build_flow_status_snapshot,
    ensure_worker,
    issue_md_path,
    seed_issue_from_github,
    seed_issue_from_text,
)
from ....core.flows.worker_process import (
    FlowWorkerHealth,
    check_worker_health,
    write_worker_exit_info,
)
from ....core.runtime import RuntimeContext
from ....core.safe_paths import SafePathError, validate_single_filename
from ....core.utils import atomic_write, find_repo_root
from ....flows.ticket_flow import build_ticket_flow_definition
from ....integrations.agents.build_agent_pool import build_agent_pool
from ....integrations.github.service import GitHubError, GitHubService
from ....tickets.bulk import bulk_clear_model_pin, bulk_set_agent
from ....tickets.files import (
    list_ticket_paths,
    parse_ticket_index,
    read_ticket,
    safe_relpath,
)
from ....tickets.frontmatter import parse_markdown_frontmatter
from ....tickets.lint import lint_ticket_directory, lint_ticket_frontmatter
from ....tickets.outbox import parse_dispatch, resolve_outbox_paths
from ..schemas import (
    TicketBulkClearModelRequest,
    TicketBulkSetAgentRequest,
    TicketBulkUpdateResponse,
    TicketCreateRequest,
    TicketDeleteResponse,
    TicketReorderRequest,
    TicketReorderResponse,
    TicketResponse,
    TicketUpdateRequest,
)

_logger = logging.getLogger(__name__)

_supported_flow_types = ("ticket_flow",)
_FLOW_DB_CORRUPT_SUFFIX = ".corrupt"
_FLOW_DB_NOTICE_SUFFIX = ".corrupt.json"


@dataclass
class FlowRoutesState:
    active_workers: Dict[
        str, Tuple[Optional[subprocess.Popen], Optional[IO[bytes]], Optional[IO[bytes]]]
    ]
    controller_cache: Dict[tuple[Path, str], FlowController]
    definition_cache: Dict[tuple[Path, str], FlowDefinition]
    lock: threading.Lock

    def __init__(self) -> None:
        self.active_workers = {}
        self.controller_cache = {}
        self.definition_cache = {}
        self.lock = threading.Lock()


def _utc_stamp() -> str:
    return datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%SZ")


def _is_probably_corrupt_flow_db_error(exc: Exception, db_path: Path) -> bool:
    if not isinstance(exc, sqlite3.Error):
        return False
    msg = str(exc).lower()
    if "file is not a database" in msg or "database disk image is malformed" in msg:
        return True
    if "disk i/o error" in msg:
        try:
            header = db_path.read_bytes()[:16]
        except Exception:
            return False
        return header not in (b"", b"SQLite format 3\x00")
    return False


def _rotate_corrupt_flow_db(db_path: Path, detail: str) -> Optional[Path]:
    stamp = _utc_stamp()
    backup_path = db_path.with_name(f"{db_path.name}{_FLOW_DB_CORRUPT_SUFFIX}.{stamp}")
    notice_path = db_path.with_name(f"{db_path.name}{_FLOW_DB_NOTICE_SUFFIX}")
    db_path.parent.mkdir(parents=True, exist_ok=True)
    backup_value: str = ""
    if db_path.exists():
        try:
            db_path.replace(backup_path)
            backup_value = str(backup_path)
        except Exception:
            backup_value = ""

    for suffix in ("-wal", "-shm"):
        sidecar = db_path.with_name(f"{db_path.name}{suffix}")
        try:
            sidecar.unlink(missing_ok=True)
        except Exception:
            pass

    notice = {
        "status": "corrupt",
        "message": "Flow store reset due to corrupted flows.db.",
        "detail": detail,
        "detected_at": stamp,
        "backup_path": backup_value,
    }
    try:
        atomic_write(notice_path, json.dumps(notice, indent=2) + "\n")
    except Exception:
        _logger.warning("Failed to write flow DB corruption notice at %s", notice_path)
    return backup_path if backup_value else None


def _evict_cached_controller(
    repo_root: Path, flow_type: str, state: FlowRoutesState
) -> None:
    key = (repo_root.resolve(), flow_type)
    with state.lock:
        controller = state.controller_cache.pop(key, None)
    if not controller:
        return
    try:
        controller.shutdown()
    except Exception:
        pass


def _recover_flow_store_if_possible(
    repo_root: Path,
    flow_type: str,
    state: FlowRoutesState,
    exc: Exception,
) -> bool:
    db_path, _ = _flow_paths(repo_root)
    if not _is_probably_corrupt_flow_db_error(exc, db_path):
        return False

    backup_path = _rotate_corrupt_flow_db(db_path, str(exc))
    _evict_cached_controller(repo_root, flow_type, state)
    store = FlowStore(db_path)
    try:
        store.initialize()
        _logger.warning(
            "Recovered corrupted flow DB at %s (backup=%s, reason=%s)",
            db_path,
            str(backup_path) if backup_path else "unavailable",
            exc,
        )
        return True
    except Exception as recover_exc:
        _logger.warning(
            "Flow DB recovery failed at %s after error %s: %s",
            db_path,
            exc,
            recover_exc,
        )
        return False
    finally:
        try:
            store.close()
        except Exception:
            pass


def _flow_paths(repo_root: Path) -> tuple[Path, Path]:
    repo_root = repo_root.resolve()
    db_path = repo_root / ".codex-autorunner" / "flows.db"
    artifacts_root = repo_root / ".codex-autorunner" / "flows"
    return db_path, artifacts_root


def _ticket_dir(repo_root: Path) -> Path:
    repo_root = repo_root.resolve()
    return repo_root / ".codex-autorunner" / "tickets"


def _find_ticket_path_by_index(ticket_dir: Path, index: int) -> Optional[Path]:
    for path in list_ticket_paths(ticket_dir):
        idx = parse_ticket_index(path.name)
        if idx == index:
            return path
    return None


_TICKET_NAME_RE = re.compile(r"^TICKET-(\d+)([^/]*)\.md$", re.IGNORECASE)


def _ticket_suffix(path: Path) -> str:
    match = _TICKET_NAME_RE.match(path.name)
    if not match:
        return ""
    return match.group(2) or ""


def _planned_ticket_renames(order: list[Path]) -> list[tuple[Path, Path]]:
    if not order:
        return []
    existing_width = max(
        3,
        max(
            (
                len(m.group(1))
                for m in (_TICKET_NAME_RE.match(path.name) for path in order)
                if m is not None
            ),
            default=3,
        ),
    )
    planned: list[tuple[Path, Path]] = []
    for new_idx, original_path in enumerate(order, start=1):
        suffix = _ticket_suffix(original_path)
        target_name = f"TICKET-{new_idx:0{existing_width}d}{suffix}.md"
        target_path = original_path.with_name(target_name)
        planned.append((original_path, target_path))
    return planned


def _rename_ticket_order(order: list[Path]) -> None:
    planned = _planned_ticket_renames(order)
    if not planned:
        return
    staged: list[tuple[Path, Path]] = []
    for original_path, _target_path in planned:
        tmp = original_path.with_name(
            f".tmp-reorder-{uuid.uuid4().hex}-{original_path.name}"
        )
        original_path.rename(tmp)
        staged.append((tmp, _target_path))
    for tmp_path, target_path in staged:
        tmp_path.rename(target_path)


def _sync_active_run_current_ticket_paths_after_reorder(
    repo_root: Path, renamed_paths: list[tuple[Path, Path]]
) -> None:
    if not renamed_paths:
        return
    store = _require_flow_store(repo_root)
    if store is None:
        return
    try:
        records = store.list_flow_runs(flow_type="ticket_flow")
        active = _active_or_paused_run(records)
        if active is None or not isinstance(active.state, dict):
            return
        rel_map: dict[str, str] = {}
        for old_path, new_path in renamed_paths:
            rel_map[safe_relpath(old_path, repo_root)] = safe_relpath(
                new_path, repo_root
            )
        next_state = dict(active.state)
        changed = False

        current_ticket = next_state.get("current_ticket")
        if isinstance(current_ticket, str):
            updated = rel_map.get(current_ticket)
            if isinstance(updated, str) and updated != current_ticket:
                next_state["current_ticket"] = updated
                changed = True

        ticket_engine = next_state.get("ticket_engine")
        if isinstance(ticket_engine, dict):
            next_ticket_engine = dict(ticket_engine)
            current_ticket = next_ticket_engine.get("current_ticket")
            if isinstance(current_ticket, str):
                updated = rel_map.get(current_ticket)
                if isinstance(updated, str) and updated != current_ticket:
                    next_ticket_engine["current_ticket"] = updated
                    next_state["ticket_engine"] = next_ticket_engine
                    changed = True

        if not changed:
            return
        store.update_flow_run_status(active.id, active.status, state=next_state)
    except Exception as exc:
        _logger.warning("Failed to sync current_ticket after reorder: %s", exc)
    finally:
        try:
            store.close()
        except Exception:
            pass


def _require_flow_store(repo_root: Path) -> Optional[FlowStore]:
    db_path, _ = _flow_paths(repo_root)
    store = FlowStore(db_path)
    try:
        store.initialize()
        return store
    except Exception as exc:
        _logger.warning("Flows database unavailable at %s: %s", db_path, exc)
        return None


def _safe_list_flow_runs(
    repo_root: Path, flow_type: Optional[str] = None, *, recover_stuck: bool = False
) -> list[FlowRunRecord]:
    db_path, _ = _flow_paths(repo_root)
    store = FlowStore(db_path)
    try:
        store.initialize()
        records = store.list_flow_runs(flow_type=flow_type)
        if recover_stuck:
            # Recover any flows stuck in active states with dead workers
            records = [
                reconcile_flow_run(repo_root, rec, store, logger=_logger)[0]
                for rec in records
            ]
        return records
    except Exception as exc:
        _logger.debug("FlowStore list runs failed: %s", exc)
        return []
    finally:
        try:
            store.close()
        except Exception:
            pass


def _build_flow_definition(
    repo_root: Path, flow_type: str, state: FlowRoutesState
) -> FlowDefinition:
    repo_root = repo_root.resolve()
    key = (repo_root, flow_type)
    with state.lock:
        if key in state.definition_cache:
            return state.definition_cache[key]

    if flow_type == "ticket_flow":
        config = load_repo_config(repo_root)
        engine = RuntimeContext(
            repo_root=repo_root,
            config=config,
        )
        agent_pool = build_agent_pool(engine.config)
        definition = build_ticket_flow_definition(agent_pool=agent_pool)
    else:
        raise HTTPException(status_code=404, detail=f"Unknown flow type: {flow_type}")

    definition.validate()
    with state.lock:
        state.definition_cache[key] = definition
    return definition


def _get_flow_controller(
    repo_root: Path, flow_type: str, state: FlowRoutesState
) -> FlowController:
    repo_root = repo_root.resolve()
    key = (repo_root, flow_type)
    with state.lock:
        cached = state.controller_cache.get(key)
    if cached is not None:
        try:
            cached.initialize()
            return cached
        except Exception as exc:
            if not _recover_flow_store_if_possible(repo_root, flow_type, state, exc):
                _evict_cached_controller(repo_root, flow_type, state)
                _logger.warning("Failed to initialize cached flow controller: %s", exc)
                raise HTTPException(
                    status_code=503,
                    detail="Flows unavailable; initialize the repo first.",
                ) from exc

    db_path, artifacts_root = _flow_paths(repo_root)
    definition = _build_flow_definition(repo_root, flow_type, state)

    def _new_controller() -> FlowController:
        return FlowController(
            definition=definition,
            db_path=db_path,
            artifacts_root=artifacts_root,
        )

    controller = _new_controller()
    try:
        controller.initialize()
    except Exception as exc:
        if _recover_flow_store_if_possible(repo_root, flow_type, state, exc):
            controller = _new_controller()
            try:
                controller.initialize()
            except Exception as retry_exc:
                _logger.warning(
                    "Failed to initialize flow controller after recovery: %s", retry_exc
                )
                raise HTTPException(
                    status_code=503,
                    detail="Flows unavailable; initialize the repo first.",
                ) from retry_exc
        else:
            _logger.warning("Failed to initialize flow controller: %s", exc)
            raise HTTPException(
                status_code=503, detail="Flows unavailable; initialize the repo first."
            ) from exc
    with state.lock:
        state.controller_cache[key] = controller
    return controller


def _get_flow_record(repo_root: Path, run_id: str) -> FlowRunRecord:
    store = _require_flow_store(repo_root)
    if store is None:
        raise HTTPException(status_code=503, detail="Flows database unavailable")
    try:
        record = store.get_flow_run(run_id)
    except sqlite3.Error as exc:
        raise HTTPException(
            status_code=503, detail="Flows database unavailable"
        ) from exc
    finally:
        try:
            store.close()
        except Exception:
            pass
    if not record:
        raise HTTPException(status_code=404, detail=f"Flow run {run_id} not found")
    return record


def _active_or_paused_run(records: list[FlowRunRecord]) -> Optional[FlowRunRecord]:
    for record in records:
        if record.status in (FlowRunStatus.RUNNING, FlowRunStatus.PAUSED):
            return record
    return None


def _latest_run(records: list[FlowRunRecord]) -> Optional[FlowRunRecord]:
    if not records:
        return None
    return records[0]


def _normalize_run_id(run_id: Union[str, uuid.UUID]) -> str:
    try:
        return str(uuid.UUID(str(run_id)))
    except ValueError:
        raise HTTPException(status_code=400, detail="Invalid run_id") from None


def _validate_tickets(ticket_dir: Path) -> list[str]:
    """Validate all tickets in the directory and return a list of error messages."""
    errors: list[str] = []

    if not ticket_dir.exists():
        return errors

    # Check for directory-level errors (duplicate indices)
    dir_errors = lint_ticket_directory(ticket_dir)
    errors.extend(dir_errors)

    # Check each ticket file for frontmatter errors
    ticket_paths = list_ticket_paths(ticket_dir)
    for path in ticket_paths:
        _, ticket_errors = read_ticket(path)
        for err in ticket_errors:
            errors.append(f"{path.relative_to(path.parent.parent)}: {err}")

    return errors


def _lint_after_ticket_update(ticket_dir: Path) -> list[str]:
    if not ticket_dir.exists():
        return []
    return _validate_tickets(ticket_dir)


def _cleanup_worker_handle(run_id: str, state: FlowRoutesState) -> None:
    with state.lock:
        handle = state.active_workers.pop(run_id, None)
    if not handle:
        return

    proc, stdout, stderr = handle
    if proc and proc.poll() is not None:
        try:
            write_worker_exit_info(
                find_repo_root(), run_id, returncode=getattr(proc, "returncode", None)
            )
        except Exception:
            pass
    if proc and proc.poll() is None:
        try:
            proc.terminate()
        except Exception:
            pass

    for stream in (stdout, stderr):
        if stream and not stream.closed:
            try:
                stream.flush()
            except Exception:
                pass
            try:
                stream.close()
            except Exception:
                pass


def _reap_dead_worker(run_id: str, state: FlowRoutesState) -> None:
    with state.lock:
        handle = state.active_workers.get(run_id)
    if not handle:
        return
    proc, *_ = handle
    if proc and proc.poll() is not None:
        _cleanup_worker_handle(run_id, state)


class FlowStartRequest(BaseModel):
    input_data: Dict = Field(default_factory=dict)
    metadata: Optional[Dict] = None


class BootstrapCheckResponse(BaseModel):
    status: str
    github_available: Optional[bool] = None
    repo: Optional[str] = None


class SeedIssueRequest(BaseModel):
    issue_ref: Optional[str] = None  # GitHub issue number, #num, or URL
    plan_text: Optional[str] = None  # Freeform plan text when GitHub unavailable


class FlowWorkerHealthResponse(BaseModel):
    status: str
    pid: Optional[int]
    is_alive: bool
    message: Optional[str] = None
    exit_code: Optional[int] = None
    stderr_tail: Optional[str] = None

    @classmethod
    def from_health(cls, health: FlowWorkerHealth) -> "FlowWorkerHealthResponse":
        return cls(
            status=health.status,
            pid=health.pid,
            is_alive=health.is_alive,
            message=health.message,
            exit_code=getattr(health, "exit_code", None),
            stderr_tail=getattr(health, "stderr_tail", None),
        )


class FlowStatusResponse(BaseModel):
    id: str
    flow_type: str
    status: str
    current_step: Optional[str]
    created_at: str
    started_at: Optional[str]
    finished_at: Optional[str]
    error_message: Optional[str]
    state: Dict = Field(default_factory=dict)
    reason_summary: Optional[str] = None
    ticket_progress: Optional[Dict[str, int]] = None
    last_event_seq: Optional[int] = None
    last_event_at: Optional[str] = None
    worker_health: Optional[FlowWorkerHealthResponse] = None

    @classmethod
    def from_record(
        cls,
        record: FlowRunRecord,
        *,
        last_event_seq: Optional[int] = None,
        last_event_at: Optional[str] = None,
        worker_health: Optional[FlowWorkerHealth] = None,
    ) -> "FlowStatusResponse":
        state = record.state or {}
        reason_summary = None
        if isinstance(state, dict):
            value = state.get("reason_summary")
            if isinstance(value, str):
                reason_summary = value
        return cls(
            id=record.id,
            flow_type=record.flow_type,
            status=record.status.value,
            current_step=record.current_step,
            created_at=record.created_at,
            started_at=record.started_at,
            finished_at=record.finished_at,
            error_message=record.error_message,
            state=state,
            reason_summary=reason_summary,
            last_event_seq=last_event_seq,
            last_event_at=last_event_at,
            worker_health=(
                FlowWorkerHealthResponse.from_health(worker_health)
                if worker_health
                else None
            ),
        )


class FlowArtifactInfo(BaseModel):
    id: str
    kind: str
    path: str
    created_at: str
    metadata: Dict = Field(default_factory=dict)


def _build_flow_status_response(
    record: FlowRunRecord,
    repo_root: Path,
    *,
    store: Optional[FlowStore] = None,
) -> FlowStatusResponse:
    snapshot = build_flow_status_snapshot(repo_root, record, store)
    resp = FlowStatusResponse.from_record(
        record,
        last_event_seq=snapshot["last_event_seq"],
        last_event_at=snapshot["last_event_at"],
        worker_health=snapshot["worker_health"],
    )
    resp.ticket_progress = snapshot.get("ticket_progress")
    if snapshot.get("state") is not None:
        resp.state = snapshot["state"]
    return resp


def _start_flow_worker(
    repo_root: Path, run_id: str, state: FlowRoutesState
) -> Optional[subprocess.Popen]:
    normalized_run_id = _normalize_run_id(run_id)

    _reap_dead_worker(normalized_run_id, state)
    result: dict = ensure_worker(repo_root, normalized_run_id)
    if result["status"] == "reused":
        health = result["health"]
        _logger.info(
            "Worker already active for run %s (pid=%s), skipping spawn",
            normalized_run_id,
            health.pid,
        )
        return None
    proc: subprocess.Popen = result["proc"]
    stdout_handle = result["stdout"]
    stderr_handle = result["stderr"]
    with state.lock:
        state.active_workers[normalized_run_id] = (proc, stdout_handle, stderr_handle)
    _logger.info("Started flow worker for run %s (pid=%d)", normalized_run_id, proc.pid)
    return proc


def _stop_worker(run_id: str, state: FlowRoutesState, timeout: float = 10.0) -> None:
    normalized_run_id = _normalize_run_id(run_id)
    with state.lock:
        handle = state.active_workers.get(normalized_run_id)
    if not handle:
        health = check_worker_health(find_repo_root(), normalized_run_id)
        if health.is_alive and health.pid:
            try:
                _logger.info(
                    "Stopping untracked worker for run %s (pid=%s)",
                    normalized_run_id,
                    health.pid,
                )
                subprocess.run(["kill", str(health.pid)], check=False)
            except Exception as exc:
                _logger.warning(
                    "Failed to stop untracked worker %s: %s", normalized_run_id, exc
                )
        return

    proc, *_ = handle
    if proc and proc.poll() is None:
        proc.terminate()
        try:
            proc.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            _logger.warning(
                "Worker for run %s did not exit in time, killing", normalized_run_id
            )
            proc.kill()
        except Exception as exc:
            _logger.warning("Error stopping worker %s: %s", normalized_run_id, exc)

    _cleanup_worker_handle(normalized_run_id, state)


def build_flow_routes() -> APIRouter:
    router = APIRouter(prefix="/api/flows", tags=["flows"])

    state = FlowRoutesState()

    def _ensure_state_in_app(request: Request) -> FlowRoutesState:
        from typing import cast

        if not hasattr(request.app.state, "flow_routes_state"):
            request.app.state.flow_routes_state = state
        return cast(FlowRoutesState, request.app.state.flow_routes_state)

    def _definition_info(definition: FlowDefinition) -> Dict:
        return {
            "type": definition.flow_type,
            "name": definition.name,
            "description": definition.description,
            "input_schema": definition.input_schema or {},
        }

    def _resolve_outbox_for_record(record: FlowRunRecord, repo_root: Path):
        workspace_root = Path(record.input_data.get("workspace_root") or repo_root)
        runs_dir = Path(record.input_data.get("runs_dir") or ".codex-autorunner/runs")
        return resolve_outbox_paths(
            workspace_root=workspace_root, runs_dir=runs_dir, run_id=record.id
        )

    def _get_diff_stats_by_dispatch_seq(
        repo_root: Path, *, run_id: str
    ) -> dict[int, dict[str, int]]:
        """Return mapping of dispatch_seq -> diff stats for the run."""
        store = _require_flow_store(repo_root)
        if store is None:
            return {}
        try:
            events = store.get_events_by_type(run_id, FlowEventType.DIFF_UPDATED)
        except Exception:
            events = []
        finally:
            try:
                store.close()
            except Exception:
                pass

        by_seq: dict[int, dict[str, int]] = {}
        for ev in events:
            data = ev.data or {}
            try:
                seq_val = int(data.get("dispatch_seq") or 0)
            except Exception:
                continue
            if seq_val <= 0:
                continue
            by_seq[seq_val] = {
                "insertions": int(data.get("insertions") or 0),
                "deletions": int(data.get("deletions") or 0),
                "files_changed": int(data.get("files_changed") or 0),
            }
        return by_seq

    @router.get("")
    async def list_flow_definitions(request: Request):
        state = _ensure_state_in_app(request)
        repo_root = find_repo_root()
        definitions = [
            _definition_info(_build_flow_definition(repo_root, flow_type, state))
            for flow_type in _supported_flow_types
        ]
        return {"definitions": definitions}

    @router.get("/runs", response_model=list[FlowStatusResponse])
    async def list_runs(
        request: Request, flow_type: Optional[str] = None, reconcile: bool = False
    ):
        _ensure_state_in_app(request)
        repo_root = find_repo_root()
        store = _require_flow_store(repo_root)
        records: list[FlowRunRecord] = []
        try:
            if store:
                records = store.list_flow_runs(flow_type=flow_type)
                if reconcile:
                    records = [
                        reconcile_flow_run(repo_root, rec, store, logger=_logger)[0]
                        for rec in records
                    ]
            else:
                records = _safe_list_flow_runs(
                    repo_root, flow_type=flow_type, recover_stuck=reconcile
                )
            return [
                _build_flow_status_response(rec, repo_root, store=store)
                for rec in records
            ]
        finally:
            if store:
                store.close()

    @router.get("/{flow_type}")
    async def get_flow_definition(request: Request, flow_type: str):
        state = _ensure_state_in_app(request)
        repo_root = find_repo_root()
        if flow_type not in _supported_flow_types:
            raise HTTPException(
                status_code=404, detail=f"Unknown flow type: {flow_type}"
            )
        definition = _build_flow_definition(repo_root, flow_type, state)
        return _definition_info(definition)

    async def _start_flow(
        flow_type: str,
        request: FlowStartRequest,
        state: FlowRoutesState,
        *,
        force_new: bool = False,
        validate_tickets: bool = True,
    ) -> FlowStatusResponse:
        if flow_type not in _supported_flow_types:
            raise HTTPException(
                status_code=404, detail=f"Unknown flow type: {flow_type}"
            )

        repo_root = find_repo_root()
        controller = _get_flow_controller(repo_root, flow_type, state)

        if flow_type == "ticket_flow" and validate_tickets:
            ticket_dir = repo_root / ".codex-autorunner" / "tickets"
            if force_new and not list_ticket_paths(ticket_dir):
                raise HTTPException(
                    status_code=400,
                    detail=(
                        "No tickets found under .codex-autorunner/tickets. "
                        "Use /api/flows/ticket_flow/bootstrap to seed tickets."
                    ),
                )

            lint_errors = _validate_tickets(ticket_dir)
            if lint_errors:
                raise HTTPException(
                    status_code=400,
                    detail={
                        "message": "Ticket validation failed",
                        "errors": lint_errors,
                    },
                )

        # Reuse an active/paused run unless force_new is requested.
        if not force_new:
            runs = _safe_list_flow_runs(
                repo_root, flow_type=flow_type, recover_stuck=True
            )
            active = _active_or_paused_run(runs)
            if active:
                _reap_dead_worker(active.id, state)
                _start_flow_worker(repo_root, active.id, state)
                store = _require_flow_store(repo_root)
                try:
                    response = _build_flow_status_response(
                        active, repo_root, store=store
                    )
                finally:
                    if store:
                        store.close()
                response.state = response.state or {}
                response.state["hint"] = "active_run_reused"
                return response

        run_id = _normalize_run_id(uuid.uuid4())

        try:
            record = await controller.start_flow(
                input_data=request.input_data,
                run_id=run_id,
                metadata=request.metadata,
            )
        except Exception as exc:
            if _recover_flow_store_if_possible(repo_root, flow_type, state, exc):
                controller = _get_flow_controller(repo_root, flow_type, state)
                run_id = _normalize_run_id(uuid.uuid4())
                try:
                    record = await controller.start_flow(
                        input_data=request.input_data,
                        run_id=run_id,
                        metadata=request.metadata,
                    )
                except sqlite3.Error as retry_exc:
                    raise HTTPException(
                        status_code=503, detail="Flows database unavailable"
                    ) from retry_exc
            elif isinstance(exc, sqlite3.Error):
                raise HTTPException(
                    status_code=503, detail="Flows database unavailable"
                ) from exc
            else:
                raise

        _start_flow_worker(repo_root, run_id, state)

        store = _require_flow_store(repo_root)
        try:
            return _build_flow_status_response(record, repo_root, store=store)
        finally:
            if store:
                store.close()

    @router.post("/{flow_type}/start", response_model=FlowStatusResponse)
    async def start_flow(request: Request, flow_type: str, req: FlowStartRequest):
        state = _ensure_state_in_app(request)
        meta = req.metadata if isinstance(req.metadata, dict) else {}
        force_new = bool(meta.get("force_new"))
        return await _start_flow(flow_type, req, state, force_new=force_new)

    @router.get("/ticket_flow/bootstrap-check", response_model=BootstrapCheckResponse)
    async def bootstrap_check():
        """
        Determine whether ISSUE.md already exists and whether GitHub is available
        for fetching an issue before bootstrapping the ticket flow.
        """
        repo_root = find_repo_root()
        result = ux_bootstrap_check(repo_root, github_service_factory=GitHubService)
        if result.status == "ready":
            return BootstrapCheckResponse(status="ready")
        return BootstrapCheckResponse(
            status=result.status,
            github_available=result.github_available,
            repo=result.repo_slug,
        )

    @router.post("/ticket_flow/seed-issue")
    async def seed_issue(request: SeedIssueRequest):
        """Create .codex-autorunner/ISSUE.md from GitHub issue or user-provided text."""
        repo_root = find_repo_root()
        issue_path = issue_md_path(repo_root)
        issue_path.parent.mkdir(parents=True, exist_ok=True)

        # GitHub-backed path
        if request.issue_ref:
            try:
                seed = seed_issue_from_github(
                    repo_root,
                    request.issue_ref,
                    github_service_factory=GitHubService,  # type: ignore[arg-type]
                )
                atomic_write(issue_path, seed.content)
                return {
                    "status": "ok",
                    "source": "github",
                    "issue_number": seed.issue_number,
                    "repo": seed.repo_slug,
                }
            except GitHubError as exc:
                raise HTTPException(
                    status_code=exc.status_code, detail=str(exc)
                ) from exc
            except RuntimeError as exc:
                raise HTTPException(status_code=400, detail=str(exc)) from exc
            except Exception as exc:  # pragma: no cover - defensive
                _logger.exception("Failed to seed ISSUE.md from GitHub: %s", exc)
                raise HTTPException(
                    status_code=500, detail="Failed to fetch issue from GitHub"
                ) from exc

        # Manual text path
        if request.plan_text:
            content = seed_issue_from_text(request.plan_text)
            atomic_write(issue_path, content)
            return {"status": "ok", "source": "user_input"}

        raise HTTPException(
            status_code=400,
            detail="issue_ref or plan_text is required to seed ISSUE.md",
        )

    @router.post("/ticket_flow/bootstrap", response_model=FlowStatusResponse)
    async def bootstrap_ticket_flow(
        http_request: Request, request: Optional[FlowStartRequest] = None
    ):
        state = _ensure_state_in_app(http_request)
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_dir.mkdir(parents=True, exist_ok=True)
        ticket_path = ticket_dir / "TICKET-001.md"
        existing_tickets = list_ticket_paths(ticket_dir)
        tickets_exist = bool(existing_tickets)
        flow_request = request or FlowStartRequest()
        meta = flow_request.metadata if isinstance(flow_request.metadata, dict) else {}
        force_new = bool(meta.get("force_new"))

        if not force_new:
            records = _safe_list_flow_runs(
                repo_root, flow_type="ticket_flow", recover_stuck=True
            )
            active = _active_or_paused_run(records)
            if active:
                # Validate tickets before reusing active run
                lint_errors = _validate_tickets(ticket_dir)
                if lint_errors:
                    raise HTTPException(
                        status_code=400,
                        detail={
                            "message": "Ticket validation failed",
                            "errors": lint_errors,
                        },
                    )
                _reap_dead_worker(active.id, state)
                _start_flow_worker(repo_root, active.id, state)
                store = _require_flow_store(repo_root)
                try:
                    resp = _build_flow_status_response(active, repo_root, store=store)
                finally:
                    if store:
                        store.close()
                resp.state = resp.state or {}
                resp.state["hint"] = "active_run_reused"
                return resp

        seeded = False
        if not tickets_exist and not ticket_path.exists():
            bootstrap_ticket_id = f"tkt_{uuid.uuid4().hex}"
            template = f"""---
agent: codex
done: false
ticket_id: "{bootstrap_ticket_id}"
title: Bootstrap ticket plan
goal: Capture scope and seed follow-up tickets
---

You are the first ticket in a new ticket_flow run.

- Read `.codex-autorunner/ISSUE.md`. If it is missing:
  - If GitHub is available, ask the user for the issue/PR URL or number and create `.codex-autorunner/ISSUE.md` from it.
  - If GitHub is not available, write `DISPATCH.md` with `mode: pause` asking the user to describe the work (or share a doc). After the reply, create `.codex-autorunner/ISSUE.md` with their input.
- If helpful, create or update contextspace docs under `.codex-autorunner/contextspace/`:
  - `active_context.md` for current context and links
  - `decisions.md` for decisions/rationale
  - `spec.md` for requirements and constraints
- Break the work into additional `TICKET-00X.md` files with clear owners/goals; keep this ticket open until they exist.
- Place any supporting artifacts in `.codex-autorunner/runs/<run_id>/dispatch/` if needed.
- Write `DISPATCH.md` to dispatch a message to the user:
  - Use `mode: pause` (handoff) to wait for user response. This pauses execution.
  - Use `mode: notify` (informational) to message the user but keep running.
"""
            ticket_path.write_text(template, encoding="utf-8")
            seeded = True

        meta = flow_request.metadata if isinstance(flow_request.metadata, dict) else {}
        payload = FlowStartRequest(
            input_data=flow_request.input_data,
            metadata=meta | {"seeded_ticket": seeded},
        )
        validate_tickets = not tickets_exist or force_new
        return await _start_flow(
            "ticket_flow",
            payload,
            state,
            force_new=force_new,
            validate_tickets=validate_tickets,
        )

    @router.get("/ticket_flow/tickets")
    async def list_ticket_files():
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        # Compute cumulative diff stats per ticket for the latest ticket_flow run.
        # This keeps ticket-level +/- counters visible after a run reaches COMPLETED.
        runs = _safe_list_flow_runs(
            repo_root, flow_type="ticket_flow", recover_stuck=True
        )
        latest_run = _latest_run(runs)
        diff_by_ticket: dict[str, dict[str, int]] = {}
        if latest_run:
            store = _require_flow_store(repo_root)
            if store is not None:
                try:
                    events = store.get_events_by_type(
                        latest_run.id, FlowEventType.DIFF_UPDATED
                    )
                except Exception:
                    events = []
                finally:
                    try:
                        store.close()
                    except Exception:
                        pass
                for ev in events:
                    data = ev.data or {}
                    ticket_id = data.get("ticket_id")
                    if not isinstance(ticket_id, str) or not ticket_id.strip():
                        continue
                    stats = diff_by_ticket.setdefault(
                        ticket_id,
                        {"insertions": 0, "deletions": 0, "files_changed": 0},
                    )
                    stats["insertions"] += int(data.get("insertions") or 0)
                    stats["deletions"] += int(data.get("deletions") or 0)
                    stats["files_changed"] += int(data.get("files_changed") or 0)

        tickets = []
        for path in list_ticket_paths(ticket_dir):
            doc, errors = read_ticket(path)
            idx = getattr(doc, "index", None) or parse_ticket_index(path.name)
            # When frontmatter is broken, still surface the raw ticket body so
            # the user can inspect and fix the file in the UI instead of seeing
            # an empty card.
            try:
                raw_body = path.read_text(encoding="utf-8")
                _, parsed_body = parse_markdown_frontmatter(raw_body)
            except Exception:
                parsed_body = None
            rel_path = safe_relpath(path, repo_root)
            tickets.append(
                {
                    "path": rel_path,
                    "index": idx,
                    "chat_key": ticket_chat_scope(idx, path) if idx else None,
                    "frontmatter": asdict(doc.frontmatter) if doc else None,
                    "body": doc.body if doc else parsed_body,
                    "errors": errors,
                    "diff_stats": diff_by_ticket.get(rel_path),
                }
            )
        return {
            "ticket_dir": safe_relpath(ticket_dir, repo_root),
            "tickets": tickets,
        }

    @router.get("/ticket_flow/tickets/{index}", response_model=TicketResponse)
    async def get_ticket(index: int):
        """Fetch a single ticket by index; return raw body even if frontmatter is invalid."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_path = _find_ticket_path_by_index(ticket_dir, index)

        if not ticket_path:
            raise HTTPException(status_code=404, detail=f"Ticket {index:03d} not found")

        doc, errors = read_ticket(ticket_path)
        if doc and not errors:
            return TicketResponse(
                path=safe_relpath(ticket_path, repo_root),
                index=doc.index,
                chat_key=ticket_chat_scope(doc.index, ticket_path),
                frontmatter=asdict(doc.frontmatter),
                body=doc.body,
            )

        # Mirror list endpoint: surface raw body for repair when frontmatter is broken.
        try:
            raw_body = ticket_path.read_text(encoding="utf-8")
            parsed_frontmatter, parsed_body = parse_markdown_frontmatter(raw_body)
        except Exception:
            parsed_frontmatter, parsed_body = {}, None

        return TicketResponse(
            path=safe_relpath(ticket_path, repo_root),
            index=parse_ticket_index(ticket_path.name) or 0,
            chat_key=ticket_chat_scope(index, ticket_path),
            frontmatter=parsed_frontmatter or {},
            body=parsed_body or "",
        )

    @router.post("/ticket_flow/tickets", response_model=TicketResponse)
    async def create_ticket(request: TicketCreateRequest):
        """Create a new ticket with auto-generated index."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_dir.mkdir(parents=True, exist_ok=True)

        # Find next available index
        existing_paths = list_ticket_paths(ticket_dir)
        existing_indices = []
        for p in existing_paths:
            idx = parse_ticket_index(p.name)
            if idx is not None:
                existing_indices.append(idx)

        # Always append at the end of the sequence; do not backfill gaps.
        next_index = (max(existing_indices) + 1) if existing_indices else 1

        # Build frontmatter (quote scalars to avoid YAML parse issues with colons, etc.)
        def _quote(val: Optional[str]) -> str:
            return (
                json.dumps(val) if val is not None else ""
            )  # JSON string is valid YAML scalar

        title_line = f"title: {_quote(request.title)}\n" if request.title else ""
        goal_line = f"goal: {_quote(request.goal)}\n" if request.goal else ""
        ticket_id = f"tkt_{uuid.uuid4().hex}"
        ticket_id_line = f"ticket_id: {_quote(ticket_id)}\n"

        content = (
            "---\n"
            f"agent: {_quote(request.agent)}\n"
            "done: false\n"
            f"{ticket_id_line}"
            f"{title_line}"
            f"{goal_line}"
            "---\n\n"
            f"{request.body}\n"
        )

        ticket_path = ticket_dir / f"TICKET-{next_index:03d}.md"
        atomic_write(ticket_path, content)

        # Read back to validate and return
        doc, errors = read_ticket(ticket_path)
        if errors or not doc:
            raise HTTPException(
                status_code=400, detail=f"Failed to create valid ticket: {errors}"
            )

        return TicketResponse(
            path=safe_relpath(ticket_path, repo_root),
            index=doc.index,
            chat_key=ticket_chat_scope(doc.index, ticket_path),
            frontmatter=asdict(doc.frontmatter),
            body=doc.body,
        )

    @router.put("/ticket_flow/tickets/{index}", response_model=TicketResponse)
    async def update_ticket(index: int, request: TicketUpdateRequest):
        """Update an existing ticket by index."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_path = _find_ticket_path_by_index(ticket_dir, index)

        if not ticket_path:
            raise HTTPException(status_code=404, detail=f"Ticket {index:03d} not found")

        # Validate frontmatter before saving
        data, body = parse_markdown_frontmatter(request.content)
        _, errors = lint_ticket_frontmatter(data)
        if errors:
            raise HTTPException(
                status_code=400,
                detail={"message": "Invalid ticket frontmatter", "errors": errors},
            )

        atomic_write(ticket_path, request.content)

        # Read back to return validated data
        doc, read_errors = read_ticket(ticket_path)
        if read_errors or not doc:
            raise HTTPException(
                status_code=400, detail=f"Failed to save valid ticket: {read_errors}"
            )

        return TicketResponse(
            path=safe_relpath(ticket_path, repo_root),
            index=doc.index,
            chat_key=ticket_chat_scope(doc.index, ticket_path),
            frontmatter=asdict(doc.frontmatter),
            body=doc.body,
        )

    @router.delete("/ticket_flow/tickets/{index}", response_model=TicketDeleteResponse)
    async def delete_ticket(index: int):
        """Delete a ticket by index."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_path = _find_ticket_path_by_index(ticket_dir, index)

        if not ticket_path:
            raise HTTPException(status_code=404, detail=f"Ticket {index:03d} not found")

        rel_path = safe_relpath(ticket_path, repo_root)
        ticket_path.unlink()

        return TicketDeleteResponse(
            status="deleted",
            index=index,
            path=rel_path,
        )

    @router.post("/ticket_flow/tickets/reorder", response_model=TicketReorderResponse)
    async def reorder_ticket(request: TicketReorderRequest):
        """Reorder one ticket relative to another ticket."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        ticket_paths = list_ticket_paths(ticket_dir)
        indices = [parse_ticket_index(path.name) for path in ticket_paths]
        ordered_indices = [idx for idx in indices if idx is not None]

        source_index = request.source_index
        destination_index = request.destination_index
        place_after = bool(request.place_after)

        if source_index == destination_index:
            lint_errors = _lint_after_ticket_update(ticket_dir)
            return TicketReorderResponse(
                status="ok" if not lint_errors else "error",
                source_index=source_index,
                destination_index=destination_index,
                place_after=place_after,
                lint_errors=lint_errors,
            )
        if source_index not in ordered_indices:
            raise HTTPException(
                status_code=404, detail=f"Ticket {source_index:03d} not found"
            )
        if destination_index not in ordered_indices:
            raise HTTPException(
                status_code=404, detail=f"Ticket {destination_index:03d} not found"
            )

        source_pos = ordered_indices.index(source_index)
        destination_pos = ordered_indices.index(destination_index)
        desired_pos = destination_pos + (1 if place_after else 0)
        if source_pos < desired_pos:
            desired_pos -= 1
        desired_pos = max(0, min(desired_pos, len(ticket_paths) - 1))

        if desired_pos != source_pos:
            source_path = ticket_paths[source_pos]
            remaining_paths = [
                path for idx, path in enumerate(ticket_paths) if idx != source_pos
            ]
            reordered_paths = (
                remaining_paths[:desired_pos]
                + [source_path]
                + remaining_paths[desired_pos:]
            )
            renamed_paths = _planned_ticket_renames(reordered_paths)
            try:
                _rename_ticket_order(reordered_paths)
                _sync_active_run_current_ticket_paths_after_reorder(
                    repo_root, renamed_paths
                )
            except Exception as exc:
                raise HTTPException(
                    status_code=400, detail=f"Failed to reorder ticket: {exc}"
                ) from exc

        lint_errors = _lint_after_ticket_update(ticket_dir)
        return TicketReorderResponse(
            status="ok" if not lint_errors else "error",
            source_index=source_index,
            destination_index=destination_index,
            place_after=place_after,
            lint_errors=lint_errors,
        )

    @router.post(
        "/ticket_flow/tickets/bulk-set-agent", response_model=TicketBulkUpdateResponse
    )
    async def bulk_set_agent_route(request: TicketBulkSetAgentRequest):
        """Bulk set agent for tickets, optionally limited by range."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        try:
            result = bulk_set_agent(
                ticket_dir,
                request.agent,
                request.range,
                repo_root=repo_root,
            )
        except ValueError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from None

        lint_errors = _lint_after_ticket_update(ticket_dir)
        status = "ok" if not result.errors and not lint_errors else "error"
        return TicketBulkUpdateResponse(
            status=status,
            updated=result.updated,
            skipped=result.skipped,
            errors=result.errors,
            lint_errors=lint_errors,
        )

    @router.post(
        "/ticket_flow/tickets/bulk-clear-model", response_model=TicketBulkUpdateResponse
    )
    async def bulk_clear_model_route(request: TicketBulkClearModelRequest):
        """Bulk clear model/reasoning overrides for tickets, optionally limited by range."""
        repo_root = find_repo_root()
        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        try:
            result = bulk_clear_model_pin(
                ticket_dir,
                request.range,
                repo_root=repo_root,
            )
        except ValueError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from None

        lint_errors = _lint_after_ticket_update(ticket_dir)
        status = "ok" if not result.errors and not lint_errors else "error"
        return TicketBulkUpdateResponse(
            status=status,
            updated=result.updated,
            skipped=result.skipped,
            errors=result.errors,
            lint_errors=lint_errors,
        )

    @router.post("/{run_id}/stop", response_model=FlowStatusResponse)
    async def stop_flow(http_request: Request, run_id: str):
        state = _ensure_state_in_app(http_request)
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, run_id)
        controller = _get_flow_controller(repo_root, record.flow_type, state)

        _stop_worker(run_id, state)

        updated = await controller.stop_flow(run_id)
        store = _require_flow_store(repo_root)
        try:
            return _build_flow_status_response(updated, repo_root, store=store)
        finally:
            if store:
                store.close()

    @router.post("/{run_id}/resume", response_model=FlowStatusResponse)
    async def resume_flow(http_request: Request, run_id: str, force: bool = False):
        state = _ensure_state_in_app(http_request)
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, run_id)
        controller = _get_flow_controller(repo_root, record.flow_type, state)

        # Validate tickets before resuming ticket_flow
        if record.flow_type == "ticket_flow":
            ticket_dir = repo_root / ".codex-autorunner" / "tickets"
            lint_errors = _validate_tickets(ticket_dir)
            if lint_errors:
                raise HTTPException(
                    status_code=400,
                    detail={
                        "message": "Ticket validation failed",
                        "errors": lint_errors,
                    },
                )

        try:
            updated = await controller.resume_flow(run_id, force=force)
        except ValueError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        _reap_dead_worker(run_id, state)
        _start_flow_worker(repo_root, run_id, state)

        store = _require_flow_store(repo_root)
        try:
            return _build_flow_status_response(updated, repo_root, store=store)
        finally:
            if store:
                store.close()

    @router.post("/{run_id}/reconcile", response_model=FlowStatusResponse)
    async def reconcile_flow(http_request: Request, run_id: str):
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, run_id)
        store = _require_flow_store(repo_root)
        if not store:
            raise HTTPException(status_code=503, detail="Flow store unavailable")
        try:
            record = reconcile_flow_run(repo_root, record, store, logger=_logger)[0]
            return _build_flow_status_response(record, repo_root, store=store)
        finally:
            store.close()

    @router.post("/{run_id}/archive")
    async def archive_flow(
        http_request: Request,
        run_id: str,
        delete_run: bool = True,
        force: bool = False,
    ):
        """Archive a completed flow by moving tickets to the run's artifact directory.

        Args:
            run_id: The flow run to archive.
            delete_run: Whether to delete the run record after archiving.
            force: If True, allow archiving flows stuck in stopping/paused state
                   by force-stopping the worker first.
        """
        state = _ensure_state_in_app(http_request)
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, run_id)

        # Allow archiving terminal flows, or force-archiving stuck flows
        if not FlowRunStatus(record.status).is_terminal():
            if force and record.status in (
                FlowRunStatus.STOPPING,
                FlowRunStatus.PAUSED,
            ):
                # Force-stop any remaining worker before archiving
                _stop_worker(run_id, state, timeout=2.0)
                _logger.info(
                    "Force-archiving flow %s in %s state", run_id, record.status.value
                )
            else:
                raise HTTPException(
                    status_code=400,
                    detail="Can only archive completed/stopped/failed flows (use force=true for stuck flows)",
                )

        # Move tickets to run artifacts directory
        _, artifacts_root = _flow_paths(repo_root)
        archive_dir = artifacts_root / run_id / "archived_tickets"
        archive_dir.mkdir(parents=True, exist_ok=True)

        ticket_dir = repo_root / ".codex-autorunner" / "tickets"
        archived_count = 0
        for ticket_path in list_ticket_paths(ticket_dir):
            dest = archive_dir / ticket_path.name
            shutil.move(str(ticket_path), str(dest))
            archived_count += 1

        # Archive runs directory (dispatch_history, reply_history, etc.) to dismiss notifications
        outbox_paths = _resolve_outbox_for_record(record, repo_root)
        run_dir = outbox_paths.run_dir
        if run_dir.exists() and run_dir.is_dir():
            archived_runs_dir = artifacts_root / run_id / "archived_runs"
            shutil.move(str(run_dir), str(archived_runs_dir))

        # Delete run record if requested
        if delete_run:
            store = _require_flow_store(repo_root)
            if store:
                store.delete_flow_run(run_id)
                store.close()

        return {
            "status": "archived",
            "run_id": run_id,
            "tickets_archived": archived_count,
        }

    @router.get("/{run_id}/status", response_model=FlowStatusResponse)
    async def get_flow_status(
        http_request: Request, run_id: str, reconcile: bool = False
    ):
        state = _ensure_state_in_app(http_request)
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()

        _reap_dead_worker(run_id, state)

        record = _get_flow_record(repo_root, run_id)
        store = _require_flow_store(repo_root)
        try:
            if reconcile and store:
                record = reconcile_flow_run(repo_root, record, store, logger=_logger)[0]
            return _build_flow_status_response(record, repo_root, store=store)
        finally:
            if store:
                store.close()

    @router.get("/{run_id}/events")
    async def stream_flow_events(
        http_request: Request, run_id: str, after: Optional[int] = None
    ):
        state = _ensure_state_in_app(http_request)
        run_id = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, run_id)
        controller = _get_flow_controller(repo_root, record.flow_type, state)

        async def event_stream():
            try:
                resume_after = after
                if resume_after is None:
                    last_event_id = http_request.headers.get("Last-Event-ID")
                    if last_event_id:
                        try:
                            resume_after = int(last_event_id)
                        except ValueError:
                            _logger.debug(
                                "Invalid Last-Event-ID %s for run %s",
                                last_event_id,
                                run_id,
                            )
                async for event in controller.stream_events(
                    run_id, after_seq=resume_after
                ):
                    data = event.model_dump(mode="json")
                    yield f"id: {event.seq}\ndata: {json.dumps(data)}\n\n"
            except Exception as e:
                _logger.exception("Error streaming events for run %s: %s", run_id, e)
                raise

        return StreamingResponse(
            event_stream(),
            media_type="text/event-stream",
            headers={
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
                "X-Accel-Buffering": "no",
            },
        )

    @router.get("/{run_id}/dispatch_history")
    async def get_dispatch_history(http_request: Request, run_id: str):
        """Get dispatch history for a flow run.

        Returns all dispatches (agent->human communications) for this run.
        """
        normalized = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, normalized)
        paths = _resolve_outbox_for_record(record, repo_root)

        # Pull diff stats from FlowStore keyed by dispatch sequence number so we
        # can enrich dispatch history entries without relying on DISPATCH.md metadata.
        diff_by_seq = _get_diff_stats_by_dispatch_seq(repo_root, run_id=normalized)

        history_entries = []
        history_dir = paths.dispatch_history_dir
        if history_dir.exists() and history_dir.is_dir():
            for entry in sorted(
                [p for p in history_dir.iterdir() if p.is_dir()],
                key=lambda p: p.name,
                reverse=True,
            ):
                dispatch_path = entry / "DISPATCH.md"
                dispatch, errors = (
                    parse_dispatch(dispatch_path)
                    if dispatch_path.exists()
                    else (None, ["Dispatch file missing"])
                )
                dispatch_dict = asdict(dispatch) if dispatch else None
                if dispatch_dict and dispatch:
                    dispatch_dict["is_handoff"] = dispatch.is_handoff
                    # Add structured diff stats (per turn summary), matched by seq.
                    try:
                        entry_seq_int = int(entry.name)
                    except Exception:
                        entry_seq_int = 0
                    if entry_seq_int and entry_seq_int in diff_by_seq:
                        dispatch_dict["diff_stats"] = diff_by_seq[entry_seq_int]
                attachments = []
                for child in sorted(entry.rglob("*")):
                    if child.name == "DISPATCH.md":
                        continue
                    rel = child.relative_to(entry).as_posix()
                    if any(part.startswith(".") for part in Path(rel).parts):
                        continue
                    if child.is_dir():
                        continue
                    attachments.append(
                        {
                            "name": child.name,
                            "rel_path": rel,
                            "path": safe_relpath(child, repo_root),
                            "size": child.stat().st_size if child.is_file() else None,
                            "url": f"api/flows/{normalized}/dispatch_history/{entry.name}/{quote(rel)}",
                        }
                    )
                history_entries.append(
                    {
                        "seq": entry.name,
                        "dispatch": dispatch_dict,
                        "errors": errors,
                        "attachments": attachments,
                        "path": safe_relpath(entry, repo_root),
                    }
                )

        return {"run_id": normalized, "history": history_entries}

    @router.get("/{run_id}/reply_history/{seq}/{file_path:path}")
    def get_reply_history_file(run_id: str, seq: str, file_path: str):
        repo_root = find_repo_root()
        db_path, _ = _flow_paths(repo_root)
        store = FlowStore(db_path)
        try:
            store.initialize()
            record = store.get_flow_run(run_id)
        finally:
            try:
                store.close()
            except Exception:
                pass
        if not record:
            raise HTTPException(status_code=404, detail="Run not found")

        if not (len(seq) == 4 and seq.isdigit()):
            raise HTTPException(status_code=400, detail="Invalid seq")

        try:
            filename = validate_single_filename(file_path)
        except SafePathError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc

        input_data = dict(record.input_data or {})
        workspace_root = Path(input_data.get("workspace_root") or repo_root)
        runs_dir = Path(input_data.get("runs_dir") or ".codex-autorunner/runs")
        from ....tickets.replies import resolve_reply_paths

        reply_paths = resolve_reply_paths(
            workspace_root=workspace_root, runs_dir=runs_dir, run_id=run_id
        )
        target = reply_paths.reply_history_dir / seq / filename
        if not target.exists() or not target.is_file():
            raise HTTPException(status_code=404, detail="File not found")
        return FileResponse(path=str(target), filename=filename)

    @router.get("/{run_id}/dispatch_history/{seq}/{file_path:path}")
    async def get_dispatch_file(run_id: str, seq: str, file_path: str):
        """Get an attachment file from a dispatch history entry."""
        normalized = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, normalized)
        paths = _resolve_outbox_for_record(record, repo_root)

        base_history = paths.dispatch_history_dir.resolve()

        seq_clean = seq.strip()
        if not re.fullmatch(r"[0-9]{4}", seq_clean):
            raise HTTPException(
                status_code=400, detail="Invalid dispatch history sequence"
            )

        history_dir = (base_history / seq_clean).resolve()
        if not history_dir.is_relative_to(base_history) or not history_dir.is_dir():
            raise HTTPException(
                status_code=404, detail=f"Dispatch history not found for run {run_id}"
            )

        file_rel = PurePosixPath(file_path)
        if file_rel.is_absolute() or ".." in file_rel.parts or "\\" in file_path:
            raise HTTPException(status_code=400, detail="Invalid dispatch file path")

        safe_parts = [part for part in file_rel.parts if part not in {"", "."}]
        if any(not re.fullmatch(r"[A-Za-z0-9._-]+", part) for part in safe_parts):
            raise HTTPException(status_code=400, detail="Invalid dispatch file path")

        target = (history_dir / Path(*safe_parts)).resolve()
        try:
            resolved = target.resolve()
        except OSError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc

        if not resolved.exists():
            raise HTTPException(status_code=404, detail="File not found")

        if not resolved.is_relative_to(history_dir):
            raise HTTPException(
                status_code=403,
                detail="Access denied: file outside dispatch history directory",
            )

        return FileResponse(resolved, filename=resolved.name)

    @router.get("/{run_id}/artifacts", response_model=list[FlowArtifactInfo])
    async def list_flow_artifacts(http_request: Request, run_id: str):
        state = _ensure_state_in_app(http_request)
        normalized = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, normalized)
        controller = _get_flow_controller(repo_root, record.flow_type, state)

        artifacts = controller.get_artifacts(normalized)
        return [
            FlowArtifactInfo(
                id=art.id,
                kind=art.kind,
                path=art.path,
                created_at=art.created_at,
                metadata=art.metadata,
            )
            for art in artifacts
        ]

    @router.get("/{run_id}/artifact")
    async def get_flow_artifact(
        http_request: Request, run_id: str, kind: Optional[str] = None
    ):
        state = _ensure_state_in_app(http_request)
        normalized = _normalize_run_id(run_id)
        repo_root = find_repo_root()
        record = _get_flow_record(repo_root, normalized)
        controller = _get_flow_controller(repo_root, record.flow_type, state)

        artifacts_root = controller.get_artifacts_dir(normalized)
        if not artifacts_root:
            from fastapi import HTTPException

            raise HTTPException(
                status_code=404, detail=f"Artifact directory not found for run {run_id}"
            )

        artifacts = controller.get_artifacts(normalized)

        if kind:
            matching = [a for a in artifacts if a.kind == kind]
        else:
            matching = artifacts

        if not matching:
            from fastapi import HTTPException

            raise HTTPException(
                status_code=404,
                detail=f"No artifact found for run {run_id} with kind={kind}",
            )

        artifact = matching[0]
        artifact_path = Path(artifact.path)

        if not artifact_path.exists():
            from fastapi import HTTPException

            raise HTTPException(
                status_code=404, detail=f"Artifact file not found: {artifact.path}"
            )

        if not artifact_path.resolve().is_relative_to(artifacts_root.resolve()):
            from fastapi import HTTPException

            raise HTTPException(
                status_code=403,
                detail="Access denied: artifact path outside run directory",
            )

        return FileResponse(artifact_path, filename=artifact_path.name)

    return router



from pathlib import Path
from typing import Optional

import yaml
from typer.testing import CliRunner

from codex_autorunner.cli import app
from codex_autorunner.core.config import CONFIG_FILENAME
from codex_autorunner.core.git_utils import run_git
from codex_autorunner.tickets.frontmatter import parse_markdown_frontmatter


def _init_repo(repo_path: Path, *, branch: Optional[str] = None) -> str:
    repo_path.mkdir(parents=True, exist_ok=True)
    run_git(["init"], repo_path, check=True)
    run_git(["config", "user.email", "test@example.com"], repo_path, check=True)
    run_git(["config", "user.name", "Test User"], repo_path, check=True)
    if branch:
        run_git(["checkout", "-b", branch], repo_path, check=True)
    return (
        run_git(["symbolic-ref", "--short", "HEAD"], repo_path, check=True).stdout or ""
    ).strip()


def _commit_file(repo_path: Path, rel_path: str, content: str) -> tuple[str, str]:
    file_path = repo_path / rel_path
    file_path.parent.mkdir(parents=True, exist_ok=True)
    file_path.write_text(content, encoding="utf-8")
    run_git(["add", rel_path], repo_path, check=True)
    run_git(["commit", "-m", "add template"], repo_path, check=True)
    commit = (run_git(["rev-parse", "HEAD"], repo_path).stdout or "").strip()
    tree_entry = (
        run_git(["ls-tree", commit, "--", rel_path], repo_path).stdout or ""
    ).strip()
    blob_sha = tree_entry.split()[2]
    return commit, blob_sha


def _write_templates_config(
    hub_root: Path,
    *,
    enabled: bool,
    repos: list[dict],
) -> None:
    config_path = hub_root / CONFIG_FILENAME
    config = yaml.safe_load(config_path.read_text(encoding="utf-8"))
    config["templates"] = {"enabled": enabled, "repos": repos}
    config_path.write_text(
        yaml.safe_dump(config, sort_keys=False),
        encoding="utf-8",
    )


def test_templates_apply_next_index_writes_file(hub_env, tmp_path: Path) -> None:
    repo_path = tmp_path / "templates_repo"
    branch = _init_repo(repo_path)
    content = "---\nagent: codex\ndone: false\n---\n\n# Template\nHello\n"
    _commit_file(repo_path, "tickets/TICKET-REVIEW.md", content)

    _write_templates_config(
        hub_env.hub_root,
        enabled=True,
        repos=[
            {
                "id": "local",
                "url": str(repo_path),
                "trusted": True,
                "default_ref": branch,
            }
        ],
    )

    ticket_dir = hub_env.repo_root / ".codex-autorunner" / "tickets"
    ticket_dir.mkdir(parents=True, exist_ok=True)
    (ticket_dir / "TICKET-001.md").write_text("---\nagent: codex\ndone: false\n---\n")
    (ticket_dir / "TICKET-003.md").write_text("---\nagent: codex\ndone: false\n---\n")

    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "templates",
            "apply",
            "local:tickets/TICKET-REVIEW.md",
            "--repo",
            str(hub_env.repo_root),
        ],
    )

    assert result.exit_code == 0
    created_path = ticket_dir / "TICKET-002.md"
    assert created_path.exists()
    frontmatter, body = parse_markdown_frontmatter(
        created_path.read_text(encoding="utf-8")
    )
    assert frontmatter["agent"] == "codex"
    assert frontmatter["done"] is False
    assert frontmatter["template"] == f"local:tickets/TICKET-REVIEW.md@{branch}"
    assert frontmatter["template_commit"]
    assert frontmatter["template_blob"]
    assert "Hello" in body


def test_templates_apply_set_agent_overrides_frontmatter(
    hub_env, tmp_path: Path
) -> None:
    repo_path = tmp_path / "templates_repo"
    branch = _init_repo(repo_path)
    content = "---\nagent: opencode\ndone: false\n---\n\n# Template\nHello\n"
    _commit_file(repo_path, "tickets/TICKET-REVIEW.md", content)

    _write_templates_config(
        hub_env.hub_root,
        enabled=True,
        repos=[
            {
                "id": "local",
                "url": str(repo_path),
                "trusted": True,
                "default_ref": branch,
            }
        ],
    )

    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "templates",
            "apply",
            "local:tickets/TICKET-REVIEW.md",
            "--repo",
            str(hub_env.repo_root),
            "--set-agent",
            "user",
        ],
    )

    assert result.exit_code == 0
    ticket_dir = hub_env.repo_root / ".codex-autorunner" / "tickets"
    created_path = ticket_dir / "TICKET-001.md"
    assert created_path.exists()
    frontmatter, _body = parse_markdown_frontmatter(
        created_path.read_text(encoding="utf-8")
    )
    assert frontmatter["agent"] == "user"


def test_templates_apply_with_provenance_includes_metadata(
    hub_env, tmp_path: Path
) -> None:
    """Test that --provenance adds provenance keys to ticket frontmatter."""
    repo_path = tmp_path / "templates_repo"
    branch = _init_repo(repo_path)
    content = "---\nagent: codex\ndone: false\n---\n\n# Template\nHello\n"
    commit, blob_sha = _commit_file(repo_path, "tickets/TICKET-REVIEW.md", content)

    _write_templates_config(
        hub_env.hub_root,
        enabled=True,
        repos=[
            {
                "id": "local",
                "url": str(repo_path),
                "trusted": True,
                "default_ref": branch,
            }
        ],
    )

    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "templates",
            "apply",
            "local:tickets/TICKET-REVIEW.md",
            "--repo",
            str(hub_env.repo_root),
            "--provenance",
        ],
    )

    assert result.exit_code == 0
    ticket_dir = hub_env.repo_root / ".codex-autorunner" / "tickets"
    created_path = ticket_dir / "TICKET-001.md"
    assert created_path.exists()

    frontmatter, _body = parse_markdown_frontmatter(
        created_path.read_text(encoding="utf-8")
    )

    # Check provenance keys
    assert frontmatter["template"] == f"local:tickets/TICKET-REVIEW.md@{branch}"
    assert frontmatter["template_commit"] == commit
    assert frontmatter["template_blob"] == blob_sha
    assert frontmatter["template_trusted"] is True
    assert frontmatter["template_scan"] == "skipped"

    # Original frontmatter keys should be preserved
    assert frontmatter["agent"] == "codex"
    assert frontmatter["done"] is False


def test_templates_apply_without_provenance_no_metadata(
    hub_env, tmp_path: Path
) -> None:
    """Test that --no-provenance (default) does not add provenance keys."""
    repo_path = tmp_path / "templates_repo"
    branch = _init_repo(repo_path)
    content = "---\nagent: codex\ndone: false\n---\n\n# Template\nHello\n"
    _commit_file(repo_path, "tickets/TICKET-REVIEW.md", content)

    _write_templates_config(
        hub_env.hub_root,
        enabled=True,
        repos=[
            {
                "id": "local",
                "url": str(repo_path),
                "trusted": True,
                "default_ref": branch,
            }
        ],
    )

    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "templates",
            "apply",
            "local:tickets/TICKET-REVIEW.md",
            "--repo",
            str(hub_env.repo_root),
            "--no-provenance",
        ],
    )

    assert result.exit_code == 0
    ticket_dir = hub_env.repo_root / ".codex-autorunner" / "tickets"
    created_path = ticket_dir / "TICKET-001.md"
    assert created_path.exists()

    frontmatter, _body = parse_markdown_frontmatter(
        created_path.read_text(encoding="utf-8")
    )

    # Provenance keys should not be present
    assert "template" not in frontmatter
    assert "template_commit" not in frontmatter
    assert "template_blob" not in frontmatter
    assert "template_trusted" not in frontmatter
    assert "template_scan" not in frontmatter

    # Original frontmatter keys should be preserved
    assert frontmatter["agent"] == "codex"
    assert frontmatter["done"] is False


def test_template_apply_alias_with_out_writes_to_requested_directory(
    hub_env, tmp_path: Path
) -> None:
    repo_path = tmp_path / "templates_repo"
    branch = _init_repo(repo_path)
    content = "---\nagent: codex\ndone: false\n---\n\n# Template\nHello\n"
    _commit_file(repo_path, "tickets/TICKET-REVIEW.md", content)

    _write_templates_config(
        hub_env.hub_root,
        enabled=True,
        repos=[
            {
                "id": "local",
                "url": str(repo_path),
                "trusted": True,
                "default_ref": branch,
            }
        ],
    )

    out_dir = hub_env.repo_root / ".codex-autorunner" / "tickets-out"
    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "template",
            "apply",
            "local:tickets/TICKET-REVIEW.md",
            "--repo",
            str(hub_env.repo_root),
            "--out",
            str(out_dir),
        ],
    )

    assert result.exit_code == 0
    created_path = out_dir / "TICKET-001.md"
    assert created_path.exists()
    frontmatter, _body = parse_markdown_frontmatter(
        created_path.read_text(encoding="utf-8")
    )
    assert frontmatter["agent"] == "codex"
    assert frontmatter["done"] is False
    assert frontmatter["template"] == f"local:tickets/TICKET-REVIEW.md@{branch}"



import asyncio
import time

import pytest

from codex_autorunner.agents.opencode import runtime as opencode_runtime
from codex_autorunner.agents.opencode.events import SSEEvent
from codex_autorunner.agents.opencode.runtime import (
    collect_opencode_output,
    collect_opencode_output_from_events,
    extract_session_id,
    parse_message_response,
)


async def _iter_events(events):
    for event in events:
        yield event


def test_extract_session_id_prefers_nested_session_id() -> None:
    payload = {"session": {"id": "session-xyz"}}
    assert extract_session_id(payload) == "session-xyz"


@pytest.mark.anyio
async def test_collect_output_uses_delta() -> None:
    seen_deltas: list[str] = []

    async def _part_handler(part_type: str, part: dict[str, str], delta_text):
        if part_type == "text" and delta_text:
            seen_deltas.append(delta_text)

    events = [
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Hello "},'
            '"part":{"type":"text","text":"Hello "}}}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"world"},'
            '"part":{"type":"text","text":"Hello world"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    # Deltas are added to final output (for progress) and sent to part_handler
    assert output.text == "Hello world"
    assert seen_deltas == ["Hello ", "world"]
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_full_text_growth() -> None:
    events = [
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"part":{"id":"p1","type":"text",'
            '"text":"Hello"}}}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"part":{"id":"p1","type":"text",'
            '"text":"Hello world"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Hello world"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_session_error() -> None:
    events = [
        SSEEvent(
            event="session.error",
            data='{"sessionID":"s1","error":{"message":"boom"}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == ""
    assert output.error == "boom"


@pytest.mark.anyio
async def test_collect_output_auto_replies_question() -> None:
    replies = []

    async def _reply(request_id: str, answers: list[list[str]]) -> None:
        replies.append((request_id, answers))

    events = [
        SSEEvent(
            event="question.asked",
            data='{"sessionID":"s1","properties":{"id":"q1","questions":[{"text":"Continue?",'
            '"options":[{"label":"Yes"},{"label":"No"}]}]}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        question_policy="auto_first_option",
        reply_question=_reply,
    )
    assert output.text == ""
    assert replies == [("q1", [["Yes"]])]


@pytest.mark.anyio
async def test_collect_output_question_deduplicates() -> None:
    replies = []

    async def _reply(request_id: str, answers: list[list[str]]) -> None:
        replies.append((request_id, answers))

    events = [
        SSEEvent(
            event="question.asked",
            data='{"sessionID":"s1","properties":{"id":"q1","questions":[{"text":"Continue?",'
            '"options":[{"label":"Yes"},{"label":"No"}]}]}}',
        ),
        SSEEvent(
            event="question.asked",
            data='{"sessionID":"s1","properties":{"id":"q1","questions":[{"text":"Continue?",'
            '"options":[{"label":"Yes"},{"label":"No"}]}]}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        question_policy="auto_first_option",
        reply_question=_reply,
    )
    assert len(replies) == 1


@pytest.mark.anyio
async def test_collect_output_filters_reasoning_and_includes_legacy_none_type() -> None:
    events = [
        # Legacy text part with type=None should be included in output
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Hello "},"part":{"text":"Hello "}}}',
        ),
        # Explicit text part should be included in output
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"world"},"part":{"type":"text","text":"world"}}}',
        ),
        # Reasoning part should be excluded from output
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"thinking..."},"part":{"type":"reasoning","id":"r1","text":"thinking..."}}}',
        ),
        # Another text part with type=None should be included
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"!"},"part":{"text":"!"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    # All text content (except reasoning) should be in output
    # This tests that parts with type=None (legacy) are included
    assert output.text == "Hello world!"
    # Reasoning should be excluded
    assert "thinking" not in output.text.lower()
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_skips_reasoning_when_type_missing_on_delta() -> None:
    events = [
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"think"},"part":{"type":"reasoning","id":"r1","text":"think"}}}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":" more"},"part":{"id":"r1","text":"think more"}}}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Hello"},"part":{"type":"text","text":"Hello"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Hello"
    assert "think" not in output.text.lower()


@pytest.mark.anyio
async def test_collect_output_emits_usage_from_properties_info_tokens() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.updated",
            data=(
                '{"sessionID":"s1","properties":{"info":{"tokens":'
                '{"input":10,"output":5,"reasoning":2,"cache":{"read":1}},'
                '"modelContextWindow":2000}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert len(seen) == 1
    usage = seen[0]
    assert usage["totalTokens"] == 18
    assert usage["inputTokens"] == 10
    assert usage["cachedInputTokens"] == 1
    assert usage["outputTokens"] == 5
    assert usage["reasoningTokens"] == 2
    assert usage["modelContextWindow"] == 2000


@pytest.mark.anyio
async def test_collect_output_emits_usage_from_message_part_tokens() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.part.updated",
            data=(
                '{"sessionID":"s1","properties":{"part":{"type":"step-finish",'
                '"tokens":{"input":11,"output":4,"reasoning":1,'
                '"cache":{"read":2}},"modelContextWindow":4096}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert len(seen) == 1
    usage = seen[0]
    assert usage["totalTokens"] == 18
    assert usage["inputTokens"] == 11
    assert usage["cachedInputTokens"] == 2
    assert usage["outputTokens"] == 4
    assert usage["reasoningTokens"] == 1
    assert usage["modelContextWindow"] == 4096


@pytest.mark.anyio
async def test_collect_output_dedupes_usage_between_part_and_message_events() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.part.updated",
            data=(
                '{"sessionID":"s1","properties":{"part":{"type":"step-finish",'
                '"tokens":{"input":10,"output":5},"modelContextWindow":2000}}}'
            ),
        ),
        SSEEvent(
            event="message.updated",
            data=(
                '{"sessionID":"s1","properties":{"info":{"tokens":'
                '{"input":10,"output":5},"modelContextWindow":2000}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert len(seen) == 1


@pytest.mark.anyio
async def test_collect_output_emits_usage_when_detail_breakdown_changes() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.part.updated",
            data=(
                '{"sessionID":"s1","properties":{"part":{"type":"step-finish",'
                '"tokens":{"input":12,"output":3},"modelContextWindow":2000}}}'
            ),
        ),
        SSEEvent(
            event="message.updated",
            data=(
                '{"sessionID":"s1","properties":{"info":{"tokens":'
                '{"input":10,"output":3,"cache":{"read":2}},'
                '"modelContextWindow":2000}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert len(seen) == 2
    assert seen[0]["totalTokens"] == 15
    assert "cachedInputTokens" not in seen[0]
    assert seen[1]["totalTokens"] == 15
    assert seen[1]["cachedInputTokens"] == 2


@pytest.mark.anyio
async def test_collect_output_prefers_aggregate_usage_over_part_tokens() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.part.updated",
            data=(
                '{"sessionID":"s1","properties":{"info":{"tokens":'
                '{"input":100,"output":50}},"part":{"type":"step-finish",'
                '"tokens":{"input":1,"output":1}}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert len(seen) == 1
    assert seen[0]["totalTokens"] == 150
    assert seen[0]["inputTokens"] == 100
    assert seen[0]["outputTokens"] == 50


@pytest.mark.anyio
async def test_collect_output_backfills_context_from_providers() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.updated",
            data='{"sessionID":"s1","info":{"tokens":{"input":12,"output":3}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]

    async def _fetch_providers():
        return {
            "providers": [
                {"id": "prov", "models": {"model": {"limit": {"context": 1024}}}}
            ]
        }

    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        model_payload={"providerID": "prov", "modelID": "model"},
        part_handler=_part_handler,
        provider_fetcher=_fetch_providers,
    )
    assert output.text == ""
    assert output.error is None
    assert seen == [
        {
            "providerID": "prov",
            "modelID": "model",
            "totalTokens": 15,
            "inputTokens": 12,
            "outputTokens": 3,
            "modelContextWindow": 1024,
        }
    ]


@pytest.mark.anyio
async def test_collect_output_skips_usage_for_non_primary_session() -> None:
    seen: list[dict[str, int]] = []

    async def _part_handler(part_type: str, part: dict[str, int], delta_text):
        if part_type == "usage":
            seen.append(part)

    events = [
        SSEEvent(
            event="message.updated",
            data=(
                '{"sessionID":"s2","properties":{"info":{"tokens":'
                '{"input":3,"output":4}}}}'
            ),
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        progress_session_ids={"s1", "s2"},
        part_handler=_part_handler,
    )
    assert output.text == ""
    assert seen == []


@pytest.mark.anyio
async def test_collect_output_uses_completed_text_when_no_parts() -> None:
    events = [
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"m1","role":"assistant"},'
            '"parts":[{"type":"text","text":"Hello"}]}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Hello"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_ignores_completed_text_when_role_missing() -> None:
    events = [
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"m1","role":null},'
            '"parts":[{"type":"text","text":"User prompt"}]}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == ""
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_uses_completed_text_after_role_update() -> None:
    events = [
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"m1","role":null},'
            '"parts":[{"type":"text","text":"Hello"}]}',
        ),
        SSEEvent(
            event="message.updated",
            data='{"sessionID":"s1","info":{"id":"m1","role":"assistant"}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Hello"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_drops_user_completed_when_role_missing() -> None:
    events = [
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"u1","role":null},'
            '"parts":[{"type":"text","text":"User prompt"}]}',
        ),
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"a1","role":"assistant"},'
            '"parts":[{"type":"text","text":"Assistant response"}]}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Assistant response"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_drops_user_prompt_without_message_ids() -> None:
    events = [
        # User prompt arrives without a message id; should not be echoed.
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"User prompt"},'
            '"part":{"type":"text","text":"User prompt"}}}',
        ),
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"u1","role":"user"}}',
        ),
        # Assistant reply also lacks a message id; should be preserved.
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Assistant reply"},'
            '"part":{"type":"text","text":"Assistant reply"}}}',
        ),
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"a1","role":"assistant"}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Assistant reply"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_dedupes_completed_before_part_updates() -> None:
    events = [
        SSEEvent(
            event="message.completed",
            data='{"sessionID":"s1","info":{"id":"m1","role":"assistant"},'
            '"parts":[{"type":"text","text":"Hello"}]}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Hello"},'
            '"part":{"id":"p1","messageId":"m1","type":"text","text":"Hello"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
    )
    assert output.text == "Hello"
    assert output.error is None


@pytest.mark.anyio
async def test_collect_output_does_not_duplicate_when_final_part_update_has_no_delta() -> (
    None
):
    seen_deltas: list[str] = []

    async def _part_handler(part_type: str, part: dict[str, str], delta_text):
        if part_type == "text" and delta_text:
            seen_deltas.append(delta_text)

    events = [
        # Delta updates for a text part
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"Hello "},'
            '"part":{"id":"p1","type":"text","text":"Hello "}}}',
        ),
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"delta":{"text":"world!"},'
            '"part":{"id":"p1","type":"text","text":"Hello world!"}}}',
        ),
        # Final part update with full text, no delta (with time.end)
        SSEEvent(
            event="message.part.updated",
            data='{"sessionID":"s1","properties":{"part":{"id":"p1","type":"text",'
            '"text":"Hello world!","time":{"end":"2024-01-01T00:00:00Z"}}}',
        ),
        SSEEvent(event="session.idle", data='{"sessionID":"s1"}'),
    ]
    output = await collect_opencode_output_from_events(
        _iter_events(events),
        session_id="s1",
        part_handler=_part_handler,
    )
    # Deltas are sent to part_handler
    assert seen_deltas == ["Hello ", "world!"]
    # Final output contains the text exactly once (from deltas), not duplicated
    # The final non-delta update doesn't re-add the text because dedupe bookkeeping
    # was updated during delta processing
    assert output.text == "Hello world!"
    assert output.error is None


def test_parse_message_response() -> None:
    payload = {
        "info": {"id": "turn-1", "error": "bad auth"},
        "parts": [{"type": "text", "text": "Hello"}],
    }
    result = parse_message_response(payload)
    assert result.text == "Hello"
    assert result.error == "bad auth"


@pytest.mark.anyio
async def test_collect_output_poll_treats_missing_status_as_idle() -> None:
    """Stall recovery should treat a missing session status entry as idle and finish."""

    class _FakeClient:
        def __init__(self):
            self.session_status_calls = 0

        def stream_events(self, *, directory, ready_event=None):
            async def _gen():
                while True:
                    await asyncio.sleep(3600)
                    yield SSEEvent(event="keepalive", data="{}")

            return _gen()

        async def session_status(self, *, directory):
            self.session_status_calls += 1
            # Simulate OpenCode's sparse status map: session missing => idle
            return {}

        async def respond_permission(self, **kwargs):
            return None

        async def reply_question(self, *args, **kwargs):
            return None

        async def reject_question(self, *args, **kwargs):
            return None

        async def providers(self, **kwargs):
            return {}

    client = _FakeClient()
    start = time.monotonic()
    output = await collect_opencode_output(
        client,
        session_id="s1",
        workspace_path=".",
        stall_timeout_seconds=0.01,
    )
    elapsed = time.monotonic() - start

    # Should exit quickly via polling path and not hang indefinitely
    assert elapsed < 0.5
    assert output.text == ""
    assert output.error is None
    assert client.session_status_calls >= 1


@pytest.mark.anyio
async def test_collect_output_bounds_stall_reconnect_loop(monkeypatch) -> None:
    statuses: list[dict[str, object]] = []
    progress_events: list[dict[str, object]] = []

    async def _status_fetcher():
        statuses.append({"status": {"type": "busy"}})
        return {"status": {"type": "busy"}}

    async def _part_handler(part_type: str, part: dict[str, object], delta_text):
        if part_type == "status":
            progress_events.append(part)
        return None

    async def _never_event_stream():
        while True:
            await asyncio.sleep(3600)
            yield SSEEvent(event="keepalive", data="{}")

    monkeypatch.setattr(
        opencode_runtime,
        "_OPENCODE_STREAM_RECONNECT_BACKOFF_SECONDS",
        (0.0, 0.0),
    )
    monkeypatch.setattr(
        opencode_runtime,
        "_OPENCODE_STREAM_MAX_STALL_RECONNECT_ATTEMPTS",
        2,
    )
    monkeypatch.setattr(
        opencode_runtime,
        "_OPENCODE_STREAM_MAX_STALL_RECONNECT_SECONDS",
        999.0,
    )

    output = await collect_opencode_output_from_events(
        None,
        session_id="s1",
        event_stream_factory=lambda: _never_event_stream(),
        session_fetcher=_status_fetcher,
        part_handler=_part_handler,
        stall_timeout_seconds=0.001,
    )

    assert output.text == ""
    assert output.error is not None
    assert "opencode_stream_stalled_timeout" in output.error
    assert len(statuses) >= 1
    assert any(event.get("type") == "reconnecting" for event in progress_events)
    assert any(event.get("type") == "stall_timeout" for event in progress_events)



from __future__ import annotations

import pytest

from codex_autorunner.voice import (
    PushToTalkCapture,
    VoiceConfig,
    resolve_speech_provider,
)


class _RecorderStream:
    def __init__(self):
        self.chunks = []

    def send_chunk(self, chunk):
        self.chunks.append(chunk)
        return []

    def flush_final(self):
        return []

    def abort(self, reason=None):
        pass


class _InspectingProvider:
    name = "openai_whisper"
    supports_streaming = True

    def __init__(self):
        self.last_session = None
        self.stream = _RecorderStream()

    def start_stream(self, session):
        self.last_session = session
        return self.stream


def test_voice_config_env_overrides_and_provider_defaults():
    raw = {
        "enabled": False,
        "provider": "openai_whisper",
        "latency_mode": "quality",
        "chunk_ms": 800,
        "sample_rate": 44_100,
        "push_to_talk": {"max_ms": 10_000},
        "providers": {"openai_whisper": {"base_url": "https://api.example.com"}},
    }
    env = {
        "CODEX_AUTORUNNER_VOICE_ENABLED": "1",
        "CODEX_AUTORUNNER_VOICE_PROVIDER": "custom",
        "CODEX_AUTORUNNER_VOICE_LATENCY": "realtime",
        "CODEX_AUTORUNNER_VOICE_CHUNK_MS": "450",
        "CODEX_AUTORUNNER_VOICE_SAMPLE_RATE": "22050",
        "CODEX_AUTORUNNER_VOICE_MAX_MS": "9000",
        "CODEX_AUTORUNNER_VOICE_SILENCE_MS": "750",
        "CODEX_AUTORUNNER_VOICE_MIN_HOLD_MS": "80",
    }

    cfg = VoiceConfig.from_raw(raw, env=env)

    assert cfg.enabled is True
    assert cfg.provider == "custom"
    assert cfg.latency_mode == "realtime"
    assert cfg.chunk_ms == 450
    assert cfg.sample_rate == 22_050
    assert cfg.push_to_talk.max_ms == 9000
    assert cfg.push_to_talk.silence_auto_stop_ms == 750
    assert cfg.push_to_talk.min_hold_ms == 80

    whisper_cfg = cfg.providers["openai_whisper"]
    assert whisper_cfg["model"] == "whisper-1"  # default preserved
    assert whisper_cfg["base_url"] == "https://api.example.com"  # override applied


def test_resolve_provider_requires_enabled_and_known_provider():
    # Avoid auto-enabling when local dev env has an API key loaded.
    disabled = VoiceConfig.from_raw(
        {"enabled": False}, env={"CODEX_AUTORUNNER_VOICE_ENABLED": "0"}
    )
    with pytest.raises(ValueError):
        resolve_speech_provider(disabled)

    unknown = VoiceConfig.from_raw({"enabled": True, "provider": "does_not_exist"})
    with pytest.raises(ValueError):
        resolve_speech_provider(unknown)


def test_capture_uses_latency_mode_and_chunk_size():
    provider = _InspectingProvider()
    cfg = VoiceConfig.from_raw(
        {
            "enabled": True,
            "latency_mode": "quality",
            "chunk_ms": 900,
            "warn_on_remote_api": False,
        }
    )
    capture = PushToTalkCapture(
        provider,
        cfg,
        permission_requester=lambda: True,
    )
    capture.begin_capture()
    capture.handle_chunk(b"\x00\x01")

    assert provider.last_session is not None
    assert provider.last_session.latency_mode == "quality"
    assert provider.stream.chunks and provider.stream.chunks[0].start_ms == 0
    assert provider.stream.chunks[0].end_ms == 900


def test_voice_config_local_provider_disables_remote_warning_by_default():
    cfg = VoiceConfig.from_raw(
        {
            "enabled": True,
            "provider": "local_whisper",
        },
        env={},
    )
    assert cfg.warn_on_remote_api is False


def test_voice_config_local_alias_disables_remote_warning_by_default():
    cfg = VoiceConfig.from_raw(
        {
            "enabled": True,
            "provider": "local",
        },
        env={},
    )
    assert cfg.warn_on_remote_api is False



from pathlib import Path
from types import SimpleNamespace

from fastapi.testclient import TestClient

from codex_autorunner.bootstrap import seed_hub_files
from codex_autorunner.core.app_server_threads import (
    AppServerThreadRegistry,
    default_app_server_threads_path,
)
from codex_autorunner.integrations.app_server.event_buffer import AppServerEventBuffer
from codex_autorunner.manifest import load_manifest
from codex_autorunner.server import create_hub_app
from codex_autorunner.surfaces.web import app as web_app_module


def test_hub_app_state_includes_pma_context(hub_env) -> None:
    app = create_hub_app(hub_env.hub_root)

    assert hasattr(app.state, "app_server_threads")
    assert isinstance(app.state.app_server_threads, AppServerThreadRegistry)
    assert app.state.app_server_threads.path == default_app_server_threads_path(
        Path(hub_env.hub_root)
    )

    assert hasattr(app.state, "app_server_events")
    assert isinstance(app.state.app_server_events, AppServerEventBuffer)

    assert hasattr(app.state, "opencode_supervisor")
    assert hasattr(app.state, "opencode_prune_interval")


def test_hub_dev_mode_includes_root_repo_for_source_checkout(
    tmp_path: Path, monkeypatch
) -> None:
    hub_root = tmp_path / "car-src"
    hub_root.mkdir()
    seed_hub_files(hub_root, force=True)
    (hub_root / ".git").mkdir()
    (hub_root / "src" / "codex_autorunner").mkdir(parents=True)
    (hub_root / "src" / "codex_autorunner" / "__init__.py").write_text(
        "__version__ = 'test'\n", encoding="utf-8"
    )
    (hub_root / "Makefile").write_text("all:\n\t@true\n", encoding="utf-8")
    (hub_root / "pyproject.toml").write_text(
        '[project]\nname = "codex-autorunner"\nversion = "0.0.0"\n',
        encoding="utf-8",
    )

    monkeypatch.setenv("CAR_DEV_INCLUDE_ROOT_REPO", "1")
    app = create_hub_app(hub_root)

    manifest = load_manifest(hub_root / ".codex-autorunner" / "manifest.yml", hub_root)
    assert len(manifest.repos) == 1
    assert manifest.repos[0].path == Path(".")
    assert app.state.config.include_root_repo is True


def test_hub_lifespan_reaper_uses_config_root(hub_env, monkeypatch) -> None:
    called_roots: list[Path] = []

    def _fake_reap(root: Path):
        called_roots.append(root)
        return SimpleNamespace(killed=0, removed=0, skipped=0)

    monkeypatch.setattr(web_app_module, "reap_managed_processes", _fake_reap)
    app = create_hub_app(hub_env.hub_root)
    with TestClient(app):
        pass

    assert called_roots == [app.state.config.root]



import subprocess
from pathlib import Path

import pytest

from codex_autorunner.core.patch_utils import (
    PatchError,
    apply_patch_file,
    infer_patch_strip,
    normalize_patch_text,
    preview_patch,
)


def test_normalize_patch_text_default_target_adds_strip_prefix():
    patch_text = "\n".join(
        [
            "@@ -1 +1 @@",
            "-old",
            "+new",
        ]
    )
    normalized, targets = normalize_patch_text(
        patch_text, default_target=".codex-autorunner/contextspace/active_context.md"
    )

    assert normalized.startswith(
        "--- a/.codex-autorunner/contextspace/active_context.md\n+++ b/.codex-autorunner/contextspace/active_context.md\n"
    )
    assert infer_patch_strip(targets) == 1


def test_normalize_patch_text_apply_patch_format_infers_strip():
    patch_text = "\n".join(
        [
            "*** Begin Patch",
            "*** Update File: .codex-autorunner/contextspace/decisions.md",
            "@@ -1 +1 @@",
            "-old",
            "+new",
            "*** End Patch",
        ]
    )
    normalized, targets = normalize_patch_text(patch_text)

    assert normalized.startswith(
        "--- a/.codex-autorunner/contextspace/decisions.md\n+++ b/.codex-autorunner/contextspace/decisions.md\n"
    )
    assert infer_patch_strip(targets) == 1


def test_apply_patch_file_reports_missing_patch_binary(
    tmp_path: Path, monkeypatch
) -> None:
    def fake_run(*_args, **_kwargs):
        raise FileNotFoundError("patch missing")

    monkeypatch.setattr(subprocess, "run", fake_run)
    patch_path = tmp_path / "diff.patch"
    patch_path.write_text("", encoding="utf-8")

    with pytest.raises(PatchError, match="patch command not found"):
        apply_patch_file(
            tmp_path,
            patch_path,
            ["a/.codex-autorunner/contextspace/active_context.md"],
        )


def test_preview_patch_passes_timeout(tmp_path: Path, monkeypatch) -> None:
    def fake_run(*_args, **kwargs):
        assert kwargs.get("timeout") is not None
        return subprocess.CompletedProcess(args=["patch"], returncode=0, stdout="")

    monkeypatch.setattr(subprocess, "run", fake_run)
    result = preview_patch(
        tmp_path,
        "@@ -1 +1 @@\n-old\n+new",
        ["a/.codex-autorunner/contextspace/active_context.md"],
        base_content={".codex-autorunner/contextspace/active_context.md": "current"},
    )

    assert result[".codex-autorunner/contextspace/active_context.md"] == "current"



from fastapi.testclient import TestClient

from codex_autorunner.core import filebox
from codex_autorunner.server import create_hub_app


def test_hub_filebox_delete_removes_only_resolved_file(hub_env) -> None:
    app = create_hub_app(hub_env.hub_root)
    client = TestClient(app)

    filebox.ensure_structure(hub_env.repo_root)
    (filebox.outbox_dir(hub_env.repo_root) / "shared.txt").write_bytes(b"primary")
    legacy_topic_pending = (
        hub_env.repo_root
        / ".codex-autorunner"
        / "uploads"
        / "telegram-files"
        / "topic-1"
        / "outbox"
        / "pending"
    )
    legacy_topic_pending.mkdir(parents=True, exist_ok=True)
    (legacy_topic_pending / "shared.txt").write_bytes(b"legacy")

    resp = client.delete(f"/hub/filebox/{hub_env.repo_id}/outbox/shared.txt")
    assert resp.status_code == 200
    assert resp.json() == {"status": "ok"}
    assert not (filebox.outbox_dir(hub_env.repo_root) / "shared.txt").exists()
    assert (legacy_topic_pending / "shared.txt").exists()



from __future__ import annotations

import threading

import pytest

from codex_autorunner.agents.registry import (
    CAR_PLUGIN_API_VERSION,
    AgentDescriptor,
    _check_codex_health,
    _check_opencode_health,
    _make_codex_harness,
    _make_opencode_harness,
    get_available_agents,
    get_registered_agents,
    has_capability,
    reload_agents,
    validate_agent_id,
)


@pytest.fixture
def app_ctx():
    class MockContext:
        app_server_supervisor = object()
        app_server_events = object()
        opencode_supervisor = object()

    return MockContext()


@pytest.fixture
def app_ctx_codex_only():
    class MockContext:
        app_server_supervisor = object()
        app_server_events = object()
        opencode_supervisor = None

    return MockContext()


@pytest.fixture
def app_ctx_opencode_only():
    class MockContext:
        app_server_supervisor = None
        app_server_events = None
        opencode_supervisor = object()

    return MockContext()


@pytest.fixture
def app_ctx_missing_supervisors():
    class MockContext:
        app_server_supervisor = None
        app_server_events = None
        opencode_supervisor = None

    return MockContext()


class TestValidateAgentId:
    def test_valid_agent_ids(self):
        assert validate_agent_id("codex") == "codex"
        assert validate_agent_id("opencode") == "opencode"

    def test_case_insensitive(self):
        assert validate_agent_id("CODEX") == "codex"
        assert validate_agent_id("OPENCODE") == "opencode"
        assert validate_agent_id("CoDeX") == "codex"
        assert validate_agent_id("OpEnCoDe") == "opencode"

    def test_whitespace_trimming(self):
        assert validate_agent_id("  codex  ") == "codex"
        assert validate_agent_id("\topencode\n") == "opencode"

    def test_none_input(self):
        with pytest.raises(ValueError, match="Unknown agent"):
            validate_agent_id(None)

    def test_whitespace_only(self):
        with pytest.raises(ValueError, match="Unknown agent"):
            validate_agent_id("   ")
        with pytest.raises(ValueError, match="Unknown agent"):
            validate_agent_id("\t\n")

    def test_invalid_agent_id(self):
        with pytest.raises(ValueError, match="Unknown agent.*'invalid'"):
            validate_agent_id("invalid")
        with pytest.raises(ValueError, match="Unknown agent.*'foo'"):
            validate_agent_id("foo")


class TestHasCapability:
    def test_valid_capabilities_codex(self):
        assert has_capability("codex", "threads") is True
        assert has_capability("codex", "turns") is True
        assert has_capability("codex", "review") is True
        assert has_capability("codex", "model_listing") is True
        assert has_capability("codex", "event_streaming") is True
        assert has_capability("codex", "approvals") is True

    def test_valid_capabilities_opencode(self):
        assert has_capability("opencode", "threads") is True
        assert has_capability("opencode", "turns") is True
        assert has_capability("opencode", "review") is True
        assert has_capability("opencode", "model_listing") is True
        assert has_capability("opencode", "event_streaming") is True

    def test_nonexistent_capability(self):
        assert has_capability("codex", "invalid_capability") is False

    def test_nonexistent_agent(self):
        assert has_capability("invalid_agent", "threads") is False
        assert has_capability("invalid_agent", "invalid_capability") is False

    def test_opencode_doesnt_have_approvals(self):
        assert has_capability("opencode", "approvals") is False


class _StubEntryPoint:
    def __init__(self, obj):
        self._obj = obj
        self.group = "codex_autorunner.agent_backends"
        self.name = "stub"

    def load(self):
        return self._obj


def _run_with_entrypoints(monkeypatch, entrypoints):
    """Reload registry with supplied entry points for plugin tests."""

    def _select(_group):
        return entrypoints

    import codex_autorunner.agents.registry as registry

    monkeypatch.setattr(registry, "_select_entry_points", _select)
    registry.reload_agents()
    return registry


class TestGetRegisteredAgents:
    def test_returns_dict(self):
        agents = get_registered_agents()
        assert isinstance(agents, dict)

    def test_returns_copy(self):
        agents1 = get_registered_agents()
        agents2 = get_registered_agents()
        assert agents1 is not agents2

    def test_contains_codex_and_opencode(self):
        agents = get_registered_agents()
        assert "codex" in agents
        assert "opencode" in agents

    def test_agent_descriptor_structure(self):
        agents = get_registered_agents()
        assert isinstance(agents["codex"], AgentDescriptor)
        assert isinstance(agents["opencode"], AgentDescriptor)

    def test_agent_properties(self):
        agents = get_registered_agents()
        assert agents["codex"].id == "codex"
        assert agents["codex"].name == "Codex"
        assert agents["opencode"].id == "opencode"
        assert agents["opencode"].name == "OpenCode"


class TestGetAvailableAgents:
    def test_both_agents_available(self, app_ctx):
        available = get_available_agents(app_ctx)
        assert "codex" in available
        assert "opencode" in available
        assert len(available) == 2

    def test_codex_only_available(self, app_ctx_codex_only):
        available = get_available_agents(app_ctx_codex_only)
        assert "codex" in available
        assert "opencode" not in available
        assert len(available) == 1

    def test_opencode_only_available(self, app_ctx_opencode_only):
        available = get_available_agents(app_ctx_opencode_only)
        assert "codex" not in available
        assert "opencode" in available
        assert len(available) == 1

    def test_no_agents_available(self, app_ctx_missing_supervisors):
        available = get_available_agents(app_ctx_missing_supervisors)
        assert "codex" not in available
        assert "opencode" not in available
        assert len(available) == 0

    def test_none_context_raises_attribute_error(self):
        with pytest.raises(AttributeError):
            get_available_agents(None)

    def test_malformed_context_missing_attributes_raises(self):
        class BadContext:
            pass

        with pytest.raises(AttributeError):
            get_available_agents(BadContext())


class TestCheckCodexHealth:
    def test_healthy_context(self, app_ctx):
        assert _check_codex_health(app_ctx) is True

    def test_missing_supervisor(self, app_ctx_missing_supervisors):
        assert _check_codex_health(app_ctx_missing_supervisors) is False

    def test_none_supervisor(self):
        class NoneSupervisorContext:
            app_server_supervisor = None

        assert _check_codex_health(NoneSupervisorContext()) is False


class TestCheckOpenCodeHealth:
    def test_healthy_context(self, app_ctx):
        assert _check_opencode_health(app_ctx) is True

    def test_missing_supervisor(self, app_ctx_missing_supervisors):
        assert _check_opencode_health(app_ctx_missing_supervisors) is False

    def test_none_supervisor(self):
        class NoneSupervisorContext:
            opencode_supervisor = None

        assert _check_opencode_health(NoneSupervisorContext()) is False


class TestMakeCodexHarness:
    def test_valid_context(self, app_ctx):
        from codex_autorunner.agents.codex.harness import CodexHarness

        harness = _make_codex_harness(app_ctx)
        assert isinstance(harness, CodexHarness)

    def test_missing_supervisor_raises(self, app_ctx_missing_supervisors):
        with pytest.raises(RuntimeError, match="supervisor or events missing"):
            _make_codex_harness(app_ctx_missing_supervisors)

    def test_missing_events_raises(self):
        class ContextWithSupervisorOnly:
            app_server_supervisor = object()
            app_server_events = None

        with pytest.raises(RuntimeError, match="supervisor or events missing"):
            _make_codex_harness(ContextWithSupervisorOnly())


class TestMakeOpenCodeHarness:
    def test_valid_context(self, app_ctx):
        from codex_autorunner.agents.opencode.harness import OpenCodeHarness

        harness = _make_opencode_harness(app_ctx)
        assert isinstance(harness, OpenCodeHarness)

    def test_missing_supervisor_raises(self, app_ctx_missing_supervisors):
        with pytest.raises(RuntimeError, match="supervisor missing"):
            _make_opencode_harness(app_ctx_missing_supervisors)


class TestConcurrentAccess:
    def test_concurrent_get_registered_agents_no_exceptions(self):
        exceptions = []

        def access_agents():
            try:
                get_registered_agents()
            except Exception as e:
                exceptions.append(e)

        threads = [threading.Thread(target=access_agents) for _ in range(50)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert (
            not exceptions
        ), f"Exceptions occurred during concurrent access: {exceptions}"

    def test_concurrent_get_registered_agents_consistent_results(self):
        results = []

        def access_agents():
            results.append(get_registered_agents())

        threads = [threading.Thread(target=access_agents) for _ in range(20)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(results) == 20
        for result in results:
            assert "codex" in result
            assert "opencode" in result
            assert len(result) >= 2

    def test_concurrent_reload_and_access_no_exceptions(self):
        exceptions = []

        def reload_and_access():
            try:
                for _ in range(5):
                    get_registered_agents()
                    reload_agents()
            except Exception as e:
                exceptions.append(e)

        threads = [threading.Thread(target=reload_and_access) for _ in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert (
            not exceptions
        ), f"Exceptions occurred during concurrent reload/access: {exceptions}"


class TestPluginApiCompatibility:
    def test_accepts_older_api_version(self, monkeypatch):
        older = AgentDescriptor(
            id="older",
            name="Older",
            capabilities=frozenset(),
            make_harness=lambda ctx: None,
            plugin_api_version=CAR_PLUGIN_API_VERSION - 1,
        )
        registry = _run_with_entrypoints(monkeypatch, [_StubEntryPoint(older)])
        agents = registry.get_registered_agents()
        assert "older" in agents

    def test_rejects_newer_api_version(self, monkeypatch):
        newer = AgentDescriptor(
            id="newer",
            name="Newer",
            capabilities=frozenset(),
            make_harness=lambda ctx: None,
            plugin_api_version=CAR_PLUGIN_API_VERSION + 1,
        )
        registry = _run_with_entrypoints(monkeypatch, [_StubEntryPoint(newer)])
        agents = registry.get_registered_agents()
        assert "newer" not in agents
        # built-ins remain
        assert "codex" in agents
        assert "opencode" in agents



from __future__ import annotations

from pathlib import Path

from codex_autorunner.integrations.telegram.adapter import (
    TelegramCommand,
    parse_command,
)
from codex_autorunner.integrations.telegram.commands_registry import (
    build_command_payloads,
)
from tests.fixtures.telegram_command_helpers import (
    README_REVISIT_GUIDANCE_MIN_MODULE_THRESHOLD,
    bot_command_entity,
    make_command_spec,
)

# Helper usage: this file owns cross-cutting Telegram command invariants, so use
# shared helpers for setup and keep assertions focused on behavior contracts.


def test_contract_runtime_entity_and_fallback_parity_for_mention_validation() -> None:
    token = "/resume@x"
    entities = [bot_command_entity(token)]
    assert parse_command(f"{token} 3", entities=entities) is None
    assert parse_command(f"{token} 3") is None


def test_contract_runtime_parsing_normalizes_bot_username() -> None:
    token = "/resume@CodexBot"
    entities = [bot_command_entity(token)]
    parsed = parse_command(f"{token} 3", entities=entities, bot_username=" @codexbot ")
    assert parsed == TelegramCommand(name="resume", args="3", raw="/resume@CodexBot 3")


def test_contract_registration_normalizes_name_and_runtime_rejects_uppercase() -> None:
    specs = {"Review": make_command_spec("Review", "Review command")}
    commands, invalid = build_command_payloads(specs)

    assert invalid == []
    assert commands == [{"command": "review", "description": "Review command"}]
    assert parse_command("/Review") is None


def test_contract_registration_rejects_invalid_names() -> None:
    specs = {
        "invalid-hyphen": make_command_spec("foo-bar", "bad"),
        "invalid-whitespace": make_command_spec("foo bar", "bad"),
        "invalid-mention": make_command_spec("foo@codexbot", "bad"),
        "invalid-too-long": make_command_spec("a" * 33, "bad"),
    }
    commands, invalid = build_command_payloads(specs)
    assert commands == []
    assert invalid == ["foo-bar", "foo bar", "foo@codexbot", "a" * 33]


def test_contract_readme_revisit_threshold_policy_minimum() -> None:
    assert README_REVISIT_GUIDANCE_MIN_MODULE_THRESHOLD >= 6


def test_contract_helper_guidance_module_count_below_readme_revisit_threshold() -> None:
    tests_dir = Path(__file__).resolve().parent
    guidance_module_count = sum(
        1
        for path in tests_dir.glob("test_telegram_*.py")
        if "Helper usage:" in path.read_text(encoding="utf-8")
    )
    assert guidance_module_count < README_REVISIT_GUIDANCE_MIN_MODULE_THRESHOLD



"""Analytics summary routes.

This module aggregates run/ticket/message metadata for the analytics dashboard
without relying on legacy autorunner endpoints. It intentionally reads from the
filesystem-backed ticket_flow store and ticket files to keep the UI consistent
with the rest of the app.
"""

from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Optional

from fastapi import APIRouter

from ....core.flows.failure_diagnostics import (
    format_failure_summary,
    get_failure_payload,
)
from ....core.flows.models import FlowEventType, FlowRunRecord, FlowRunStatus
from ....core.flows.store import FlowStore
from ....core.utils import find_repo_root
from ....tickets.files import list_ticket_paths, read_ticket, ticket_is_done
from ....tickets.outbox import parse_dispatch, resolve_outbox_paths
from ....tickets.replies import resolve_reply_paths


def _flows_db_path(repo_root: Path) -> Path:
    return repo_root / ".codex-autorunner" / "flows.db"


def _select_primary_run(records: list[FlowRunRecord]) -> Optional[FlowRunRecord]:
    """Select the primary run for analytics display.

    Only considers the newest run (records[0]). If it's active or paused, return it.
    If the newest run is terminal (completed/stopped/failed), return None to show idle.
    This matches the backend's _active_or_paused_run() logic and prevents showing
    stale data from old paused runs when newer runs have completed.
    """
    if not records:
        return None
    newest = records[0]
    status = FlowRunStatus(newest.status)
    if status.is_active() or status.is_paused():
        return newest
    if status in {FlowRunStatus.FAILED, FlowRunStatus.STOPPED}:
        failure = get_failure_payload(newest)
        if failure:
            return newest
    return None


def _parse_timestamp(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        if value.endswith("Z"):
            return datetime.fromisoformat(value.replace("Z", "+00:00"))
        return datetime.fromisoformat(value)
    except ValueError:
        return None


def _duration_seconds(
    started_at: Optional[str], finished_at: Optional[str], status: str
) -> Optional[float]:
    start_dt = _parse_timestamp(started_at)
    if not start_dt:
        return None
    end_dt = _parse_timestamp(finished_at)
    if not end_dt and status in {
        FlowRunStatus.RUNNING.value,
        FlowRunStatus.PAUSED.value,
        FlowRunStatus.PENDING.value,
    }:
        end_dt = datetime.now(timezone.utc)
    if not end_dt:
        return None
    return (end_dt - start_dt).total_seconds()


def _ticket_counts(ticket_dir: Path) -> dict[str, int]:
    total = 0
    done = 0
    for path in list_ticket_paths(ticket_dir):
        total += 1
        try:
            if ticket_is_done(path):
                done += 1
        except Exception:
            # Treat unreadable/invalid tickets as not-done but still count them.
            continue
    todo = max(total - done, 0)
    return {"todo": todo, "done": done, "total": total}


def _count_history_dirs(history_dir: Path) -> int:
    if not history_dir.exists() or not history_dir.is_dir():
        return 0
    count = 0
    try:
        for child in history_dir.iterdir():
            try:
                if child.is_dir() and len(child.name) == 4 and child.name.isdigit():
                    count += 1
            except OSError:
                continue
    except OSError:
        return count
    return count


def _aggregate_diff_stats(dispatch_history_dir: Path) -> Dict[str, int]:
    """Aggregate diff stats from all turn summaries in dispatch history.

    Returns dict with insertions, deletions, files_changed totals.
    """
    totals = {"insertions": 0, "deletions": 0, "files_changed": 0}
    if not dispatch_history_dir.exists() or not dispatch_history_dir.is_dir():
        return totals

    try:
        for entry_dir in dispatch_history_dir.iterdir():
            if not entry_dir.is_dir():
                continue
            if not (len(entry_dir.name) == 4 and entry_dir.name.isdigit()):
                continue
            dispatch_path = entry_dir / "DISPATCH.md"
            if not dispatch_path.exists():
                continue
            try:
                dispatch, _errors = parse_dispatch(dispatch_path)
                if dispatch and dispatch.extra:
                    diff_stats = dispatch.extra.get("diff_stats")
                    if isinstance(diff_stats, dict):
                        totals["insertions"] += int(diff_stats.get("insertions") or 0)
                        totals["deletions"] += int(diff_stats.get("deletions") or 0)
                        totals["files_changed"] += int(
                            diff_stats.get("files_changed") or 0
                        )
            except Exception:
                continue
    except OSError:
        pass

    return totals


def _build_summary(repo_root: Path) -> Dict[str, Any]:
    from ....core.config import load_repo_config

    ticket_dir = repo_root / ".codex-autorunner" / "tickets"
    db_path = _flows_db_path(repo_root)
    records: list[FlowRunRecord] = []
    if db_path.exists():
        try:
            with FlowStore(
                db_path, durable=load_repo_config(repo_root).durable_writes
            ) as store:
                records = store.list_flow_runs(flow_type="ticket_flow")
        except Exception:
            records = []

    run_record = _select_primary_run(records)

    default_run = {
        "id": None,
        "short_id": None,
        "status": "idle",
        "started_at": None,
        "finished_at": None,
        "duration_seconds": None,
        "current_step": None,
        "created_at": None,
    }

    run_data: Dict[str, Any] = default_run
    turns: Dict[str, Any] = {
        "total": None,
        "current_ticket": None,
        "dispatches": 0,
        "replies": 0,
    }
    current_ticket: Optional[str] = None
    agent_id: Optional[str] = None

    failure_payload: Optional[dict[str, Any]] = None
    failure_summary: Optional[str] = None

    if run_record:
        run_data = {
            "id": run_record.id,
            "short_id": run_record.id.split("-")[0] if run_record.id else None,
            "status": run_record.status.value,
            "started_at": run_record.started_at,
            "finished_at": run_record.finished_at,
            "duration_seconds": _duration_seconds(
                run_record.started_at, run_record.finished_at, run_record.status.value
            ),
            "current_step": run_record.current_step,
            "created_at": run_record.created_at,
        }

        state = run_record.state if isinstance(run_record.state, dict) else {}
        ticket_state = state.get("ticket_engine") if isinstance(state, dict) else {}
        if isinstance(ticket_state, dict):
            turns["total"] = ticket_state.get("total_turns")
            turns["current_ticket"] = ticket_state.get("ticket_turns")
            current_ticket = ticket_state.get("current_ticket")
            agent_id = ticket_state.get("last_agent_id")

        workspace_value = run_record.input_data.get("workspace_root")
        workspace_root = Path(workspace_value) if workspace_value else repo_root
        runs_dir = Path(
            run_record.input_data.get("runs_dir") or ".codex-autorunner/runs"
        )
        outbox_paths = resolve_outbox_paths(
            workspace_root=workspace_root, runs_dir=runs_dir, run_id=run_record.id
        )
        reply_paths = resolve_reply_paths(
            workspace_root=workspace_root, runs_dir=runs_dir, run_id=run_record.id
        )
        turns["dispatches"] = _count_history_dirs(outbox_paths.dispatch_history_dir)
        turns["replies"] = _count_history_dirs(reply_paths.reply_history_dir)
        # Diff stats are now stored in FlowStore as DIFF_UPDATED events.
        # Fallback to legacy dispatch history parsing if FlowStore query fails.
        try:
            with FlowStore(
                db_path, durable=load_repo_config(repo_root).durable_writes
            ) as store:
                events = store.get_events_by_type(
                    run_record.id, FlowEventType.DIFF_UPDATED
                )
            totals = {"insertions": 0, "deletions": 0, "files_changed": 0}
            for ev in events:
                data = ev.data or {}
                totals["insertions"] += int(data.get("insertions") or 0)
                totals["deletions"] += int(data.get("deletions") or 0)
                totals["files_changed"] += int(data.get("files_changed") or 0)
            turns["diff_stats"] = totals
        except Exception:
            turns["diff_stats"] = _aggregate_diff_stats(
                outbox_paths.dispatch_history_dir
            )

        # If current ticket is known, read its frontmatter to pick agent id when available.
        if current_ticket:
            current_path = (workspace_root / current_ticket).resolve()
            try:
                doc, _errors = read_ticket(current_path)
                if doc and doc.frontmatter and getattr(doc.frontmatter, "agent", None):
                    agent_id = doc.frontmatter.agent
            except Exception:
                pass

        failure_payload = get_failure_payload(run_record)
        if failure_payload:
            failure_summary = format_failure_summary(failure_payload)

    ticket_counts = _ticket_counts(ticket_dir)

    return {
        "run": run_data,
        "failure": failure_payload,
        "failure_summary": failure_summary,
        "tickets": {
            "todo_count": ticket_counts["todo"],
            "done_count": ticket_counts["done"],
            "total_count": ticket_counts["total"],
            "current_ticket": current_ticket,
        },
        "turns": {
            "total": turns.get("total"),
            "current_ticket": turns.get("current_ticket"),
            "dispatches": turns.get("dispatches"),
            "replies": turns.get("replies"),
            "diff_stats": turns.get("diff_stats"),
        },
        "agent": {
            "id": agent_id,
            "model": None,
        },
    }


def build_analytics_routes() -> APIRouter:
    router = APIRouter(prefix="/api/analytics", tags=["analytics"])

    @router.get("/summary")
    def get_analytics_summary():
        repo_root = find_repo_root()
        data = _build_summary(repo_root)
        return data

    return router


__all__ = ["build_analytics_routes"]



from __future__ import annotations

import json
from types import SimpleNamespace

import httpx
import pytest

from codex_autorunner.agents.opencode.client import OpenCodeProtocolError
from codex_autorunner.agents.opencode.supervisor import OpenCodeSupervisorError
from codex_autorunner.integrations.telegram.constants import TELEGRAM_MAX_MESSAGE_LENGTH
from codex_autorunner.integrations.telegram.handlers.commands.command_utils import (
    _format_httpx_exception,
    _format_opencode_exception,
    _issue_only_link,
    _opencode_review_arguments,
)
from codex_autorunner.integrations.telegram.transport import TelegramMessageTransport


def test_issue_only_link_matches_single_link_wrappers() -> None:
    link = "https://example.com/issue/1"
    assert _issue_only_link(link, [link]) == link
    assert _issue_only_link(f"<{link}>", [link]) == link
    assert _issue_only_link(f"({link})", [link]) == link


def test_issue_only_link_ignores_non_wrapper_text() -> None:
    assert _issue_only_link("", ["https://example.com"]) is None
    assert _issue_only_link("check this", ["https://example.com"]) is None
    assert _issue_only_link("https://example.com", ["one", "two"]) is None


def test_opencode_review_arguments_reduces_known_target_types() -> None:
    assert _opencode_review_arguments({"type": "uncommittedChanges"}) == ""
    assert (
        _opencode_review_arguments({"type": "baseBranch", "branch": "feature/ci"})
        == "feature/ci"
    )
    assert _opencode_review_arguments({"type": "commit", "sha": "abc123"}) == "abc123"
    assert (
        _opencode_review_arguments({"type": "custom", "instructions": "add tests"})
        == "uncommitted\n\nadd tests"
    )
    assert (
        _opencode_review_arguments({"type": "custom", "instructions": "   "})
        == "uncommitted"
    )


def test_opencode_review_arguments_falls_back_to_json_payload() -> None:
    target = {"type": "other", "foo": "bar"}
    assert _opencode_review_arguments(target) == json.dumps(target, sort_keys=True)


def test_format_httpx_exception_uses_http_payload_detail() -> None:
    request = httpx.Request("GET", "https://example.com")
    response = httpx.Response(
        502,
        request=request,
        json={"message": "temporary outage"},
    )
    exc = httpx.HTTPStatusError("server error", request=request, response=response)
    assert _format_httpx_exception(exc) == "temporary outage"


def test_format_opencode_exception_formats_backend_unavailable() -> None:
    result = _format_opencode_exception(OpenCodeSupervisorError("service offline"))
    assert result == "OpenCode backend unavailable (service offline)."


@pytest.mark.parametrize(
    ("exc", "expected"),
    [
        (
            OpenCodeProtocolError("invalid protocol"),
            "OpenCode protocol error: invalid protocol",
        ),
        (OpenCodeProtocolError(""), "OpenCode protocol error."),
    ],
)
def test_format_opencode_exception_formats_protocol_error(
    exc: Exception, expected: str
) -> None:
    assert _format_opencode_exception(exc) == expected


class _OverflowBotStub:
    def __init__(self) -> None:
        self.send_message_calls: list[dict[str, object]] = []
        self.send_message_chunks_calls: list[dict[str, object]] = []

    async def send_message(
        self,
        chat_id: int,
        text: str,
        *,
        message_thread_id: int | None = None,
        reply_to_message_id: int | None = None,
        reply_markup: dict[str, object] | None = None,
        parse_mode: str | None = None,
    ) -> dict[str, object]:
        self.send_message_calls.append(
            {
                "chat_id": chat_id,
                "text": text,
                "thread_id": message_thread_id,
                "reply_to": reply_to_message_id,
                "reply_markup": reply_markup,
                "parse_mode": parse_mode,
            }
        )
        return {"message_id": len(self.send_message_calls)}

    async def send_message_chunks(
        self,
        chat_id: int,
        text: str,
        *,
        message_thread_id: int | None = None,
        reply_to_message_id: int | None = None,
        reply_markup: dict[str, object] | None = None,
        parse_mode: str | None = None,
    ) -> None:
        self.send_message_chunks_calls.append(
            {
                "chat_id": chat_id,
                "text": text,
                "thread_id": message_thread_id,
                "reply_to": reply_to_message_id,
                "reply_markup": reply_markup,
                "parse_mode": parse_mode,
            }
        )


class _OverflowTransportHarness(TelegramMessageTransport):
    def __init__(self, overflow_mode: str) -> None:
        self._bot = _OverflowBotStub()
        self._config = SimpleNamespace(
            parse_mode="HTML", message_overflow=overflow_mode
        )
        self.sent_documents: list[dict[str, object]] = []

    def _build_debug_prefix(
        self,
        *,
        chat_id: int,
        thread_id: int | None,
        reply_to: int | None,
    ) -> str | None:
        _ = (chat_id, thread_id, reply_to)
        return None

    def _render_message(
        self, text: str, *, parse_mode: str = "HTML"
    ) -> tuple[str, str]:
        _ = parse_mode
        return text, "HTML"

    def _prepare_outgoing_text(
        self,
        text: str,
        *,
        chat_id: int,
        thread_id: int | None = None,
        reply_to: int | None = None,
        topic_key: str | None = None,
        codex_thread_id: str | None = None,
    ) -> tuple[str, str | None]:
        _ = (chat_id, thread_id, reply_to, topic_key, codex_thread_id)
        return text, "HTML"

    async def _send_document(
        self,
        chat_id: int,
        content: bytes,
        *,
        filename: str,
        thread_id: int | None = None,
        reply_to: int | None = None,
        caption: str | None = None,
    ) -> bool:
        self.sent_documents.append(
            {
                "chat_id": chat_id,
                "content_len": len(content),
                "filename": filename,
                "thread_id": thread_id,
                "reply_to": reply_to,
                "caption": caption,
            }
        )
        return True


@pytest.mark.anyio
@pytest.mark.parametrize("overflow_mode", ["document", "split", "trim"])
async def test_long_message_overflow_modes_are_stable(overflow_mode: str) -> None:
    harness = _OverflowTransportHarness(overflow_mode)
    long_text = "x" * (TELEGRAM_MAX_MESSAGE_LENGTH + 500)

    await harness._send_message(1234, long_text, thread_id=42, reply_to=99)

    if overflow_mode == "document":
        assert len(harness.sent_documents) == 1
        assert harness.sent_documents[0]["filename"] == "response.md"
        assert harness._bot.send_message_calls == []
        assert harness._bot.send_message_chunks_calls == []
        return
    if overflow_mode == "split":
        assert len(harness._bot.send_message_calls) >= 2
        assert harness._bot.send_message_calls[0]["reply_to"] == 99
        assert all(
            call["parse_mode"] == "HTML" for call in harness._bot.send_message_calls
        )
        assert harness.sent_documents == []
        assert harness._bot.send_message_chunks_calls == []
        return
    assert len(harness._bot.send_message_chunks_calls) == 1
    payload = harness._bot.send_message_chunks_calls[0]["text"]
    assert isinstance(payload, str)
    assert len(payload) <= TELEGRAM_MAX_MESSAGE_LENGTH
    assert harness.sent_documents == []
    assert harness._bot.send_message_calls == []



from codex_autorunner.integrations.telegram.constants import (
    SHELL_OUTPUT_TRUNCATION_SUFFIX,
)
from codex_autorunner.integrations.telegram.overflow import (
    split_markdown_message,
    trim_markdown_message,
)


def test_split_markdown_message_closes_code_fences() -> None:
    text = "```python\n" + ("print('x')\n" * 40) + "```"
    chunks = split_markdown_message(text, max_len=120)

    assert len(chunks) > 1
    for chunk in chunks:
        assert len(chunk) <= 120
        assert chunk.count("<pre><code>") == chunk.count("</code></pre>")


def test_trim_markdown_message_appends_suffix() -> None:
    text = "hello " * 100
    trimmed = trim_markdown_message(text, max_len=120)

    assert len(trimmed) <= 120
    assert SHELL_OUTPUT_TRUNCATION_SUFFIX.strip() in trimmed


def test_split_markdown_message_adds_continued_indicator() -> None:
    text = "\n\n".join(
        [
            "First paragraph " * 5,
            "Second paragraph " * 5,
        ]
    )
    chunks = split_markdown_message(
        text,
        max_len=80,
        render=lambda value: value,
    )

    assert len(chunks) > 1
    total = len(chunks)
    assert f"continued (1/{total})" in chunks[0]
    assert f"continued ({total}/{total})" in chunks[-1]
    assert "Second paragraph" not in chunks[0]
